# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import urllib.parse
import warnings
from typing import Dict, List, Tuple, Any, Mapping

import numpy as np
import pyproj
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.tilingscheme import EARTH_CIRCUMFERENCE_WGS84
from xcube.core.tilingscheme import GEOGRAPHIC_CRS_NAME
from xcube.core.tilingscheme import TilingScheme
from xcube.core.tilingscheme import WEB_MERCATOR_CRS_NAME
from xcube.webapi.common.xml import Document
from xcube.webapi.common.xml import Element
from xcube.constants import CRS_CRS84
from .context import WmtsContext

WMTS_VERSION = '1.0.0'
WMTS_URL_PREFIX = f'wmts/{WMTS_VERSION}'
WMTS_TILE_FORMAT = 'image/png'

WMTS_CRS84_TMS_ID = 'WorldCRS84Quad'
WMTS_CRS84_TMS_TITLE = 'CRS84 for the World'
OGC_CRS84_URN = 'urn:ogc:def:crs:OGC:1.3:CRS84'
OGC_CRS84_WKSS_URN = 'urn:ogc:def:wkss:OGC:1.0:GoogleCRS84Quad'

WMTS_WEB_MERCATOR_TMS_ID = 'WorldWebMercatorQuad'
WMTS_WEB_MERCATOR_TMS_TITLE = 'Google Maps Compatible for the World'
OGC_WEB_MERCATOR_URN = 'urn:ogc:def:crs:EPSG::3857'
OGC_WEB_MERCATOR_WKSS_URN = 'urn:ogc:def:wkss:OGC:1.0:GoogleMapsCompatible'

_STD_PIXEL_SIZE_IN_METERS = 0.28e-3

# '/tile/%s/%s/%s/' is the pattern for
# '/tile/{ds_name}/{var_name}/{tms_id}/'
_TILE_URL_TEMPLATE = WMTS_URL_PREFIX \
                     + '/tile/%s/%s/%s/' \
                       '{TileMatrix}/' \
                       '{TileRow}/' \
                       '{TileCol}.png'


def get_wmts_capabilities_xml(ctx: WmtsContext,
                              base_url: str,
                              tms_id: str = WMTS_CRS84_TMS_ID) -> str:
    """
    Get WMTSCapabilities.xml according to
    https://www.ogc.org/standards/wmts, WMTS 1.0.0.

    We create WMTSCapabilities.xml for individual tile matrix sets.
    If we'd include it into one, we would have to double all layers
    and make them available for the supported tile matrix sets
    "WorldCRS84Quad" and "WorldWebMercatorQuad".

    :param ctx: server context
    :param base_url: the request base URL
    :param tms_id: time matrix set identifier,
        must be one of "WorldCRS84Quad" or "WorldWebMercatorQuad"
    :return: XML plain text in UTF-8 encoding
    """
    element = get_capabilities_element(ctx, base_url, tms_id)
    document = Document(element)
    return document.to_xml(indent=4)


def get_capabilities_element(ctx: WmtsContext,
                             base_url: str,
                             tms_id: str) -> Element:
    service_provider_element = get_service_provider_element(ctx)
    operations_metadata_element = get_operations_metadata_element(
        ctx, base_url, tms_id
    )

    layer_base_url = ctx.get_service_url(base_url, _TILE_URL_TEMPLATE)

    contents_element = Element('Contents')
    themes_element = Element('Themes')

    common_tiling_scheme = TilingScheme.for_crs(
        get_crs_name_from_tms_id(tms_id)
    )

    for dataset_config in ctx.datasets_ctx.get_dataset_configs():
        ds_name = dataset_config['Identifier']

        ml_dataset = ctx.datasets_ctx.get_ml_dataset(ds_name)
        grid_mapping = ml_dataset.grid_mapping
        ds = ml_dataset.base_dataset

        x_name, y_name = grid_mapping.xy_dim_names

        try:
            crs84_bbox = get_crs84_bbox(grid_mapping)
        except ValueError:
            warnings.warn(f'cannot compute geographical'
                          f' bounds for dataset {ds_name}, ignoring it')
            continue

        ds_theme_element = get_ds_theme_element(ds_name, ds, dataset_config)
        themes_element.add(ds_theme_element)

        dim_elements_cache = dict()

        var_names = sorted(ds.data_vars)
        for var_name in var_names:
            var = ds[var_name]

            is_spatial_var = var.ndim >= 2 \
                and var.dims[-1] == x_name \
                and var.dims[-2] == y_name
            if not is_spatial_var:
                continue

            var_layer_element, var_theme_element = \
                get_var_layer_and_theme_element(
                    ds_name,
                    var_name,
                    var,
                    crs84_bbox,
                    layer_base_url,
                    tms_id
                )

            dim_elements = get_dim_elements(ds_name, ds, var,
                                            dim_elements_cache)

            var_layer_element.add(*dim_elements)
            contents_element.add(var_layer_element)
            ds_theme_element.add(var_theme_element)

        # Here we compute min/max level for all datasets
        # which is actually not right. Otherwise, we had to create a new
        # TMS for every dataset (or even variable).
        tiling_scheme = ml_dataset.derive_tiling_scheme(common_tiling_scheme)
        if common_tiling_scheme.min_level is None:
            min_level = tiling_scheme.min_level
            max_level = tiling_scheme.max_level
        else:
            min_level = min(common_tiling_scheme.min_level,
                            tiling_scheme.min_level)
            max_level = max(common_tiling_scheme.max_level,
                            tiling_scheme.max_level)
        common_tiling_scheme = common_tiling_scheme.derive(
            min_level=min_level, max_level=max_level
        )

    contents_element.add(
        get_tile_matrix_set_element(common_tiling_scheme),
    )

    service_identification_element = get_service_identification_element()
    service_metadata_url_element = get_service_metadata_url_element(
        ctx, base_url, tms_id
    )

    return Element(
        'Capabilities',
        attrs={
            'xmlns': "http://www.opengis.net/wmts/1.0",
            'xmlns:ows': "http://www.opengis.net/ows/1.1",
            'xmlns:xlink': "http://www.w3.org/1999/xlink",
            'xmlns:xsi': "http://www.w3.org/2001/XMLSchema-instance",
            'xsi:schemaLocation': "http://www.opengis.net/wmts/1.0"
                                  " http://schemas.opengis.net/wmts/1.0.0/"
                                  "wmtsGetCapabilities_response.xsd",
            'version': WMTS_VERSION,
        },
        elements=[
            service_identification_element,
            service_provider_element,
            operations_metadata_element,
            contents_element,
            themes_element,
            service_metadata_url_element
        ]
    )


def get_dim_elements(
        ds_name: str,
        ds: xr.Dataset,
        var: xr.DataArray,
        dim_element_cache: Dict[str, Element]
) -> List[Element]:
    dim_elements = []
    for dim_name in var.dims[0:-2]:
        if dim_name not in ds.coords:
            continue

        dim_id = f'{ds_name}.{dim_name}'
        if dim_id in dim_element_cache:
            dim_elements.append(dim_element_cache[dim_id])
            continue

        coord_var = ds.coords[dim_name]
        if len(coord_var.shape) != 1:
            # strange case
            continue

        coord_bnds_var_name = coord_var.attrs.get(
            'bounds', f'{dim_name}_bnds'
        )
        coord_bnds_var = None
        if coord_bnds_var_name in ds:
            coord_bnds_var = ds.coords[coord_bnds_var_name]
        if coord_bnds_var is not None:
            if len(coord_bnds_var.shape) != 2 \
                    or coord_bnds_var.shape[0] != \
                    coord_bnds_var.shape[0] \
                    or coord_bnds_var.shape[1] != 2:
                # strange case
                coord_bnds_var = None

        dim_title = coord_var.attrs.get('long_name', dim_name)
        if dim_name == 'time':
            units = 'ISO8601'
            default = 'current'
            current = 'true'
        else:
            units = coord_var.attrs.get('units', '')
            default = '0'
            current = 'false'

        dim_element = Element('Dimension', elements=[
            Element('ows:Identifier', text=f'{dim_name}'),
            Element('ows:Title', text=dim_title),
            Element('ows:UOM', text=units),
            Element('Default', text=default),
            Element('Current', text=current),
        ])

        if coord_bnds_var is not None:
            coord_bnds_var_values = coord_bnds_var.values
            for i in range(len(coord_var)):
                value1 = coord_bnds_var_values[i, 0]
                value2 = coord_bnds_var_values[i, 1]
                dim_element.add(
                    Element('Value', text=f'{value1}/{value2}')
                )
        else:
            coord_var_values = coord_var.values
            for i in range(len(coord_var)):
                value = coord_var_values[i]
                dim_element.add(
                    Element('Value', text=f'{value}')
                )

        dim_element_cache[dim_id] = dim_element
        dim_elements.append(dim_element)

    return dim_elements


def get_ds_theme_element(ds_name: str,
                         ds: xr.Dataset,
                         dataset_config: Mapping[str, Any]) -> Element:
    ds_title = dataset_config.get(
        'Title', ds.attrs.get('title', ds_name)
    )
    ds_abstract = dataset_config.get(
        'Abstract', ds.attrs.get('abstract',
                                 ds.attrs.get('comment', ''))
    )
    return Element('Theme', elements=[
        Element('ows:Identifier', text=ds_name),
        Element('ows:Title', text=ds_title),
        Element('ows:Abstract', text=ds_abstract),
    ])


def get_var_layer_and_theme_element(
        ds_name: str,
        var_name: str,
        var: xr.DataArray,
        var_geo_bbox: Tuple[float, float, float, float],
        var_tile_url_templ_pattern: str,
        tms_id: str
) -> Tuple[Element, Element]:
    var_id = f'{ds_name}.{var_name}'
    var_title = ds_name + "/" + var.attrs.get(
        'title', var.attrs.get('long_name', var_name)
    )
    var_abstract = var.attrs.get(
        'comment', var.attrs.get('abstract', '')
    )
    var_theme_element = Element('Theme', elements=[
        Element('ows:Identifier', text=var_id),
        Element('ows:Title', text=var_title),
        Element('ows:Abstract', text=var_abstract),
        Element('LayerRef', text=var_id),
    ])
    # noinspection PyTypeChecker
    var_tile_url_templ_params = tuple(map(urllib.parse.quote_plus,
                                          [ds_name, var_name, tms_id]))
    var_tile_url_templ = (var_tile_url_templ_pattern
                          % var_tile_url_templ_params)
    layer_element = Element('Layer', elements=[
        Element('ows:Identifier', text=f'{ds_name}.{var_name}'),
        Element('ows:Title', text=f'{var_title}'),
        Element('ows:Abstract', text=f'{var_abstract}'),
        Element('ows:WGS84BoundingBox', elements=[
            Element('ows:LowerCorner', text=f'{var_geo_bbox[0]}'
                                            f' {var_geo_bbox[1]}'),
            Element('ows:UpperCorner', text=f'{var_geo_bbox[2]}'
                                            f' {var_geo_bbox[3]}'),
        ]),
        Element('Style', attrs={'isDefault': 'true'},
                elements=[
                    Element('ows:Identifier', text='Default'),
                ]),
        Element('Format', text=WMTS_TILE_FORMAT),
        Element('TileMatrixSetLink', elements=[
            Element('TileMatrixSet', text=tms_id),
        ]),
        Element('ResourceURL',
                attrs=dict(format=WMTS_TILE_FORMAT,
                           resourceType='tile',
                           template=var_tile_url_templ)),
    ])
    return layer_element, var_theme_element


def get_dataset_tiling_scheme(ml_dataset: MultiLevelDataset, tms_id: str) \
        -> TilingScheme:
    if tms_id == WMTS_CRS84_TMS_ID:
        tiling_scheme = TilingScheme.GEOGRAPHIC
    else:
        tiling_scheme = TilingScheme.WEB_MERCATOR
    return ml_dataset.derive_tiling_scheme(tiling_scheme)


def get_tile_matrix_set_element(tiling_scheme: TilingScheme) -> Element:
    if tiling_scheme.crs_name == tiling_scheme.GEOGRAPHIC.crs_name:
        return get_tile_matrix_set_crs84_element(tiling_scheme)
    else:
        return get_tile_matrix_set_web_mercator_element(tiling_scheme)


def get_tile_matrix_set_crs84_element(tiling_scheme: TilingScheme) \
        -> Element:
    return _get_tile_matrix_set_element(
        WMTS_CRS84_TMS_ID,
        WMTS_CRS84_TMS_TITLE,
        OGC_CRS84_URN,
        OGC_CRS84_WKSS_URN,
        (-180, -90,
         +180, +90),
        EARTH_CIRCUMFERENCE_WGS84 / 360.0,
        tiling_scheme.min_level,
        tiling_scheme.max_level,
    )


def get_tile_matrix_set_web_mercator_element(tiling_scheme: TilingScheme) \
        -> Element:
    return _get_tile_matrix_set_element(
        WMTS_WEB_MERCATOR_TMS_ID,
        WMTS_WEB_MERCATOR_TMS_TITLE,
        OGC_WEB_MERCATOR_URN,
        OGC_WEB_MERCATOR_WKSS_URN,
        (-20037508.3427892, -20037508.3427892,
         +20037508.3427892, +20037508.3427892),
        1.0,
        tiling_scheme.min_level,
        tiling_scheme.max_level,
    )


def _get_tile_matrix_set_element(
        tms_id: str,
        tms_title: str,
        crs_urn: str,
        wkss_urn: str,
        bbox: Tuple[float, float, float, float],
        meters_per_pixel: float,
        min_level: int,
        max_level: int,
) -> Element:
    element = Element('TileMatrixSet', elements=[
        Element('ows:Identifier', text=tms_id),
        Element('ows:Title', text=tms_title),
        Element('ows:SupportedCRS', text=crs_urn),
        Element('ows:BoundingBox',
                attrs=dict(crs=crs_urn),
                elements=[
                    Element('ows:LowerCorner', text=f'{bbox[0]} {bbox[1]}'),
                    Element('ows:UpperCorner', text=f'{bbox[2]} {bbox[3]}')
                ]),
        Element('WellKnownScaleSet', text=wkss_urn)
    ])

    scale_factor = meters_per_pixel / _STD_PIXEL_SIZE_IN_METERS

    tiling_scheme = TilingScheme.for_crs(get_crs_name_from_tms_id(tms_id))
    tile_width = tiling_scheme.tile_width
    tile_height = tiling_scheme.tile_height
    num_x_tiles_0, num_y_tiles_0 = tiling_scheme.num_level_zero_tiles
    resolutions = tiling_scheme.get_resolutions(
        min_level=min_level,
        max_level=max_level
    )
    for level, res in zip(range(min_level, max_level + 1), resolutions):
        factor = 2 ** level
        num_x_tiles = factor * num_x_tiles_0
        num_y_tiles = factor * num_y_tiles_0
        scale_denominator = scale_factor * res
        element.add(
            Element(
                'TileMatrix',
                elements=[
                    Element('ows:Identifier', text=f'{level}'),
                    Element('ScaleDenominator', text=f'{scale_denominator}'),
                    Element('TopLeftCorner', text=f'{bbox[0]} {bbox[3]}'),
                    Element('TileWidth', text=f'{tile_width}'),
                    Element('TileHeight', text=f'{tile_height}'),
                    Element('MatrixWidth', text=f'{num_x_tiles}'),
                    Element('MatrixHeight', text=f'{num_y_tiles}'),
                ]
            )
        )

    return element


def get_operations_metadata_element(ctx: WmtsContext,
                                    base_url: str,
                                    tms_id: str) -> Element:
    wmts_kvp_url = ctx.get_service_url(
        base_url, 'wmts/kvp?'
    )
    wmts_rest_cap_url = ctx.get_service_url(
        base_url, f'{WMTS_URL_PREFIX}/{tms_id}/WMTSCapabilities.xml'
    )
    wmts_rest_tile_url = ctx.get_service_url(
        base_url, WMTS_URL_PREFIX + '/'
    )
    operations_metadata = Element(
        'ows:OperationsMetadata',
        elements=[
            _get_operation_element('GetCapabilities',
                                   wmts_kvp_url,
                                   wmts_rest_cap_url),
            _get_operation_element('GetTile',
                                   wmts_kvp_url,
                                   wmts_rest_tile_url)
        ])
    return operations_metadata


def _get_operation_element(op: str,
                           kvp_url: str,
                           rest_url: str) -> Element:
    return Element(
        'ows:Operation',
        attrs={'name': op},
        elements=[
            Element(
                'ows:DCP',
                elements=[
                    Element(
                        'ows:HTTP',
                        elements=[
                            _get_operation_get_element(kvp_url, 'KVP'),
                            _get_operation_get_element(rest_url, 'REST'),
                        ]),
                ]),
        ])


def _get_operation_get_element(url: str, value: str) -> Element:
    return Element(
        'ows:Get',
        attrs={'xlink:href': url},
        elements=[
            Element(
                'ows:Constraint',
                attrs={'name': 'GetEncoding'},
                elements=[
                    Element(
                        'ows:AllowedValues',
                        elements=[
                            Element('ows:Value', text=value)
                        ]
                    ),
                ]
            ),
        ]
    )


def get_service_provider_element(ctx):
    service_provider = ctx.config.get('ServiceProvider')

    def _get_value(path):
        v = None
        node = service_provider
        for k in path:
            if not isinstance(node, dict) or k not in node:
                return ''
            v = node[k]
            node = v
        return str(v) if v is not None else ''

    element = Element('ows:ServiceProvider', elements=[
        Element('ows:ProviderName',
                text=_get_value(['ProviderName'])),
        Element('ows:ProviderSite',
                attrs={'xlink:href': _get_value(['ProviderSite'])}),
        Element('ows:ServiceContact', elements=[
            Element('ows:IndividualName',
                    text=_get_value(['ServiceContact',
                                     'IndividualName'])),
            Element('ows:PositionName',
                    text=_get_value(['ServiceContact',
                                     'PositionName'])),
            Element('ows:ContactInfo', elements=[
                Element('ows:Phone', elements=[
                    Element('ows:Voice',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Phone',
                                             'Voice'])),
                    Element('ows:Facsimile',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Phone',
                                             'Facsimile'])),
                ]),
                Element('ows:Address', elements=[
                    Element('ows:DeliveryPoint',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Address',
                                             'DeliveryPoint'])),
                    Element('ows:City',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Address',
                                             'City'])),
                    Element('ows:AdministrativeArea',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Address',
                                             'AdministrativeArea'])),
                    Element('ows:PostalCode',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Address',
                                             'PostalCode'])),
                    Element('ows:Country',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Address',
                                             'Country'])),
                    Element('ows:ElectronicMailAddress',
                            text=_get_value(['ServiceContact',
                                             'ContactInfo',
                                             'Address',
                                             'ElectronicMailAddress'])),
                ]),
            ]),
        ]),
    ])
    return element


def get_service_identification_element():
    return Element('ows:ServiceIdentification', elements=[
        Element('ows:Title', text='xcube WMTS'),
        Element('ows:Abstract', text='Web Map Tile Service (WMTS)'
                                     ' for xcube-conformant data cubes'),
        Element('ows:Keywords',
                elements=[Element('ows:Keyword', text='tile'),
                          Element('ows:Keyword', text='tile matrix set'),
                          Element('ows:Keyword', text='map')]),
        Element('ows:ServiceType', text='OGC WMTS'),
        Element('ows:ServiceTypeVersion', text=WMTS_VERSION),
        Element('ows:Fees', text='none'),
        Element('ows:AccessConstraints', text='none'),
    ])


def get_service_metadata_url_element(
        ctx: WmtsContext,
        base_url: str,
        tms_id: str
) -> Element:
    get_caps_rest_url = ctx.get_service_url(
        base_url, f'{WMTS_URL_PREFIX}/{tms_id}/WMTSCapabilities.xml'
    )
    return Element('ServiceMetadataURL',
                   attrs={'xlink:href': get_caps_rest_url})


# TODO: move get_crs84_bbox() into GridMapping so we can adjust
#   global dataset attributes with result
def get_crs84_bbox(grid_mapping: GridMapping) \
        -> Tuple[float, float, float, float]:
    if grid_mapping.crs.is_geographic:
        return grid_mapping.xy_bbox
    t = pyproj.Transformer.from_crs(grid_mapping.crs, CRS_CRS84,
                                    always_xy=True)
    x1, y1, x2, y2 = grid_mapping.xy_bbox
    # TODO (forman): Fixme, this will be wrong
    #   for datasets crossing anti-meridian.
    x, y = t.transform(np.array([x1, x2, x1, x2]),
                       np.array([y1, y1, y2, y2]))
    coords = np.array([np.nanmin(x), np.nanmin(y),
                       np.nanmax(x), np.nanmax(y)])
    if not np.all(np.isfinite(coords)):
        raise ValueError('grid mapping bbox cannot'
                         ' be represented in geographical coordinates')
    # noinspection PyTypeChecker
    return tuple(map(float, coords))


def get_crs_name_from_tms_id(tms_id: str) -> str:
    if tms_id == WMTS_CRS84_TMS_ID:
        return GEOGRAPHIC_CRS_NAME
    if tms_id == WMTS_WEB_MERCATOR_TMS_ID:
        return WEB_MERCATOR_CRS_NAME
    raise ValueError(
        f'Tile matrix set must be one of'
        f' "{WMTS_CRS84_TMS_ID}" and "{WMTS_WEB_MERCATOR_TMS_ID}"'
    )



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import atexit
import io
import os
import shutil
import tempfile
from typing import Tuple

DEFAULT_TEMP_FILE_PREFIX = 'xcube-'


def new_temp_dir(prefix: str = None,
                 suffix: str = None,
                 dir_path: str = None) -> str:
    """
    Create new temporary directory using ``tempfile.mkdtemp()```
    The directory will be removed later when the process ends.

    :param prefix: If not ``None``, the file name will
        begin with that prefix; otherwise, a default prefix is used.
    :param suffix: If not ``None``, the file name will
        end with that suffix, otherwise there will be no suffix.
    :param dir_path: If not ``None``, the file will be created
        in that directory; otherwise, a default directory is used.
        The default directory is chosen from a platform-dependent
        list, but the user of the application can control the
        directory location by setting the *TMPDIR*, *TEMP* or *TMP*
        environment variables.
    :return: The absolute path to the new directory.
    """
    prefix = DEFAULT_TEMP_FILE_PREFIX if prefix is None else prefix
    return remove_dir_later(tempfile.mkdtemp(prefix=prefix,
                                             suffix=suffix,
                                             dir=dir_path))


def new_temp_file(prefix: str = None,
                  suffix: str = None,
                  dir_path: str = None,
                  text_mode: bool = False) -> Tuple[int, str]:
    """
    Create new temporary file using ``tempfile.mkstemp()```.
    The file will be removed later when the process ends.

    :param prefix: If not ``None``, the file name will
        begin with that prefix; otherwise, a default prefix is used.
    :param suffix: If not ``None``, the file name will
        end with that suffix, otherwise there will be no suffix.
    :param dir_path: If not ``None``, the file will be created
        in that directory; otherwise, a default directory is used.
        The default directory is chosen from a platform-dependent
        list, but the user of the application can control the
        directory location by setting the *TMPDIR*, *TEMP* or *TMP*
        environment variables.
    :param text_mode: If specified and true, the file is opened in
        text mode. Otherwise, (the default) the file is opened in
        binary mode.
    :return: A tuple comprising the file descriptor (an int) and the
        absolute path to the new file.
    """
    prefix = DEFAULT_TEMP_FILE_PREFIX if prefix is None else prefix
    fd, file_path = tempfile.mkstemp(prefix=prefix,
                                     suffix=suffix,
                                     dir=dir_path,
                                     text=text_mode)
    return fd, remove_file_later(file_path)


def remove_dir_later(dir_path: str) -> str:
    """
    Remove directory later when the process ends.
    Removal failures are ignored.

    :param dir_path: Path to directory
    :return: The absolute path to the directory.
    """
    dir_path = os.path.abspath(dir_path)

    def _remove_dir_later():
        # noinspection PyBroadException
        try:
            shutil.rmtree(dir_path, ignore_errors=True)
        except BaseException:
            pass

    atexit.register(_remove_dir_later)
    return dir_path


def remove_file_later(file_path: str) -> str:
    """
    Remove file later when the process ends.
    Removal failures are ignored.

    :param file_path: path to a file
    :return: The absolute path to the file.
    """
    file_path = os.path.abspath(file_path)

    def _remove_file_later():
        # noinspection PyBroadException
        try:
            os.remove(file_path)
        except BaseException:
            pass

    atexit.register(_remove_file_later)
    return file_path



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
import os.path
import sys
import time
from abc import ABCMeta, abstractmethod
from threading import RLock

from typing import Optional

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

from xcube.constants import LOG

_DEBUG_CACHE = False


class CacheStore(metaclass=ABCMeta):
    """
    Represents a store to which cached values can be stored into and restored from.
    """

    @abstractmethod
    def can_load_from_key(self, key) -> bool:
        """
        Test whether a stored value representation can be loaded from the given key.
        :param key: the key
        :return: True, if so
        """
        pass

    @abstractmethod
    def load_from_key(self, key):
        """
        Load a stored value representation of the value and its size from the given key.
        :param key: the key
        :return: a 2-element sequence containing the stored representation of the value and it's size
        """
        pass

    @abstractmethod
    def store_value(self, key, value):
        """
        Store a value and return it's stored representation and size in any unit, e.g. in bytes.
        :param key: the key
        :param value: the value
        :return: a 2-element sequence containing the stored representation of the value and it's size
        """
        pass

    @abstractmethod
    def restore_value(self, key, stored_value):
        """
        Restore a vale from its stored representation.
        :param key: the key
        :param stored_value: the stored representation of the value
        :return: the item
        """
        pass

    @abstractmethod
    def discard_value(self, key, stored_value):
        """
        Discard a value from it's storage.
        :param key: the key
        :param stored_value: the stored representation of the value
        """
        pass


class MemoryCacheStore(CacheStore):
    """
    Simple memory store.
    """

    def can_load_from_key(self, key) -> bool:
        # This store type does not maintain key-value pairs on its own
        return False

    def load_from_key(self, key):
        raise NotImplementedError()

    def store_value(self, key, value):
        """
        Return (value, 1).
        :param key: the key
        :param value: the original value
        :return: the tuple (stored value, size) where stored value is the sequence [key, value].
        """
        return [key, value], _compute_object_size(value)

    def restore_value(self, key, stored_value):
        """
        :param key: the key
        :param stored_value: the stored representation of the value
        :return: the original value.
        """
        if key != stored_value[0]:
            raise ValueError('key does not match stored value')
        return stored_value[1]

    def discard_value(self, key, stored_value):
        """
        Clears the value in the given stored_value.
        :param key: the key
        :param stored_value: the stored representation of the value
        """
        if key != stored_value[0]:
            raise ValueError('key does not match stored value')
        stored_value[1] = None


class FileCacheStore(CacheStore):
    """
    Simple file store for values which can be written and read as bytes, e.g. encoded PNG images.
    """

    def __init__(self, cache_dir: str, ext: str):
        self.cache_dir = cache_dir
        self.ext = ext

    def can_load_from_key(self, key) -> bool:
        path = self._key_to_path(key)
        return os.path.exists(path)

    def load_from_key(self, key):
        path = self._key_to_path(key)
        return path, os.path.getsize(path)

    def store_value(self, key, value):
        path = self._key_to_path(key)
        dir_path = os.path.dirname(path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path, exist_ok=True)
        with open(path, 'wb') as fp:
            fp.write(value)
        return path, os.path.getsize(path)

    def restore_value(self, key, stored_value):
        path = self._key_to_path(key)
        with open(path, 'rb') as fp:
            return fp.read()

    def discard_value(self, key, stored_value):
        path = self._key_to_path(key)
        try:
            os.remove(path)
            # TODO (forman): also remove empty directories up to self.cache_dir
        except IOError:
            pass

    def _key_to_path(self, key):
        return os.path.join(self.cache_dir, str(key) + self.ext)


def _policy_lru(item):
    return item.access_time


def _policy_mru(item):
    return -item.access_time


def _policy_lfu(item):
    return item.access_count


def _policy_rr(item):
    return item.access_count % 2


#: Discard Least Recently Used items first
POLICY_LRU = _policy_lru
#: Discard Most Recently Used first
POLICY_MRU = _policy_mru
#: Discard Least Frequently Used first
POLICY_LFU = _policy_lfu
#: Discard items by Random Replacement
POLICY_RR = _policy_rr

_T0 = time.process_time()


class Cache:
    """
    An implementation of a cache.
    See https://en.wikipedia.org/wiki/Cache_algorithms
    """

    class Item:
        """
        Cache-private class representing an item in the cache.
        """

        def __init__(self):
            self.key = None
            self.stored_value = None
            self.stored_size = 0
            self.creation_time = 0
            self.access_time = 0
            self.access_count = 0

        @staticmethod
        def load_from_key(store, key):
            if not store.can_load_from_key(key):
                return None
            item = Cache.Item()
            item._load_from_key(store, key)
            return item

        def store(self, store, key, value):
            self.key = key
            self.access_count = 0
            self._access()
            stored_value, stored_size = store.store_value(key, value)
            self.stored_value = stored_value
            self.stored_size = stored_size

        def restore(self, store, key):
            self._access()
            return store.restore_value(key, self.stored_value)

        def discard(self, store, key):
            store.discard_value(key, self.stored_value)
            self.__init__()

        def _load_from_key(self, store, key):
            self.key = key
            self.access_count = 0
            self._access()
            stored_value, stored_size = store.load_from_key(key)
            self.stored_value = stored_value
            self.stored_size = stored_size

        def _access(self):
            self.access_time = time.process_time() - _T0
            self.access_count += 1

    def __init__(self, store=MemoryCacheStore(), capacity=1000, threshold=0.75, policy=POLICY_LRU, parent_cache=None):
        """
        Constructor.

        :param store: the cache store, see CacheStore interface
        :param capacity: the size capacity in units used by the store's store() method
        :param threshold: a number greater than zero and less than one
        :param policy: cache replacement policy. This is a function that maps a :py:class:`Cache.Item`
                       to a numerical value. See :py:data:`POLICY_LRU`,
                       :py:data:`POLICY_MRU`, :py:data:`POLICY_LFU`, :py:data:`POLICY_RR`
        """
        self._store = store
        self._capacity = capacity
        self._threshold = threshold
        self._policy = policy
        self._parent_cache = parent_cache
        self._size = 0
        self._max_size = self._capacity * self._threshold
        self._item_dict = {}
        self._item_list = []
        self._lock = RLock()

    @property
    def policy(self):
        return self._policy

    @property
    def store(self):
        return self._store

    @property
    def capacity(self):
        return self._capacity

    @property
    def threshold(self):
        return self._threshold

    @property
    def size(self):
        return self._size

    @property
    def max_size(self):
        return self._max_size

    def get_value(self, key):
        self._lock.acquire()
        item = self._item_dict.get(key)
        value = None
        restored = False
        if item:
            value = item.restore(self._store, key)
            restored = True
            if _DEBUG_CACHE:
                _debug_print('restored value for key "%s" from cache' % key)
        elif self._parent_cache:
            item = self._parent_cache.get_value(key)
            if item:
                value = item.restore(self._parent_cache.store, key)
                restored = True
                if _DEBUG_CACHE:
                    _debug_print('restored value for key "%s" from parent cache' % key)
        if not restored:
            item = Cache.Item.load_from_key(self._store, key)
            if item:
                self._add_item(item)
                value = item.restore(self._store, key)
                if _DEBUG_CACHE:
                    _debug_print('restored value for key "%s" from cache' % key)
        self._lock.release()
        return value

    def put_value(self, key, value):
        self._lock.acquire()
        if self._parent_cache:
            # remove value from parent cache, because this cache will now take over
            self._parent_cache.remove_value(key)
        item = self._item_dict.get(key)
        if item:
            self._remove_item(item)
            item.discard(self._store, key)
            if _DEBUG_CACHE:
                _debug_print('discarded value for key "%s" from cache' % key)
        else:
            item = Cache.Item()
        item.store(self._store, key, value)
        if _DEBUG_CACHE:
            _debug_print('stored value for key "%s" in cache' % key)
        self._add_item(item)
        self._lock.release()

    def remove_value(self, key):
        self._lock.acquire()
        if self._parent_cache:
            self._parent_cache.remove_value(key)
        item = self._item_dict.get(key)
        if item:
            self._remove_item(item)
            item.discard(self._store, key)
            if _DEBUG_CACHE:
                _debug_print('Cache: discarded value for key "%s" from parent cache' % key)
        self._lock.release()

    def _add_item(self, item):
        self._item_dict[item.key] = item
        self._item_list.append(item)
        if self._size + item.stored_size > self._max_size:
            self.trim(item.stored_size)
        self._size += item.stored_size

    def _remove_item(self, item):
        self._item_dict.pop(item.key)
        self._item_list.remove(item)
        self._size -= item.stored_size

    def trim(self, extra_size=0):
        if _DEBUG_CACHE:
            _debug_print('trimming...')
        self._lock.acquire()
        self._item_list.sort(key=self._policy)
        keys = []
        size = self._size
        max_size = self._max_size
        for item in self._item_list:
            if size + extra_size > max_size:
                keys.append(item.key)
                size -= item.stored_size
        self._lock.release()
        # release lock to give another thread a chance then require lock again
        self._lock.acquire()
        for key in keys:
            if self._parent_cache:
                # Before discarding item fully, put its value into the parent cache
                value = self.get_value(key)
                self.remove_value(key)
                if value:
                    self._parent_cache.put_value(key, value)
            else:
                self.remove_value(key)
        self._lock.release()

    def clear(self, clear_parent=True):
        self._lock.acquire()
        if self._parent_cache and clear_parent:
            self._parent_cache.clear(clear_parent)
        keys = list(self._item_dict.keys())
        self._lock.release()
        for key in keys:
            if self._parent_cache and not clear_parent:
                value = self.get_value(key)
                if value:
                    self._parent_cache.put_value(key, value)
            self.remove_value(key)


def _debug_print(msg):
    LOG.debug("Cache:", msg)


def _compute_object_size(obj):
    if hasattr(obj, 'nbytes'):
        # A numpy ndarray instance
        return obj.nbytes
    elif hasattr(obj, 'size') and hasattr(obj, 'mode'):
        # A PIL Image instance
        w, h = obj.size
        m = obj.mode
        return w * h * (4 if m in ('RGBA', 'RGBx', 'I', 'F') else
                        3 if m in ('RGB', 'YCbCr', 'LAB', 'HSV') else
                        1. / 8. if m == '1' else
                        1)
    else:
        return sys.getsizeof(obj)


def parse_mem_size(mem_size_text: str) -> Optional[int]:
    mem_size_text = mem_size_text.upper()
    if mem_size_text != "" and mem_size_text not in ("OFF", "NONE", "NULL", "FALSE"):
        unit = mem_size_text[-1]
        factors = {"B": 10 ** 0, "K": 10 ** 3, "M": 10 ** 6, "G": 10 ** 9, "T": 10 ** 12}
        try:
            if unit in factors:
                capacity = int(mem_size_text[0: -1]) * factors[unit]
            else:
                capacity = int(mem_size_text)
        except ValueError:
            raise ValueError(f"invalid memory size: {mem_size_text!r}")
        if capacity > 0:
            return capacity
        elif capacity < 0:
            raise ValueError(f"negative memory size: {mem_size_text!r}")
    return None



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
from pathlib import Path
from typing import List, Optional, Any, Dict

import click

from xcube.cli.common import (cli_option_quiet,
                              cli_option_verbosity,
                              configure_cli_output)
from xcube.constants import (DEFAULT_SERVER_FRAMEWORK,
                             DEFAULT_SERVER_PORT,
                             DEFAULT_SERVER_ADDRESS)

assets_to_show = [
    'apis',
    'endpoints',
    'openapi',
    'config',
    'configschema'
]

assets_to_show_choices = list(assets_to_show)
assets_to_show_choices.extend([a + '.yaml' for a in assets_to_show])
assets_to_show_choices.extend([a + '.json' for a in assets_to_show])


@click.command(name='serve')
@click.option('--framework', 'framework_name',
              metavar='FRAMEWORK', default=DEFAULT_SERVER_FRAMEWORK,
              type=click.Choice(["tornado", "flask"]),
              help=f'Web server framework.'
                   f' Defaults to "{DEFAULT_SERVER_FRAMEWORK}"')
@click.option('--port', '-p',
              metavar='PORT', default=None, type=int,
              help=f'Service port number.'
                   f' Defaults to {DEFAULT_SERVER_PORT}')
@click.option('--address', '-a',
              metavar='ADDRESS', default=None,
              help=f'Service address.'
                   f' Defaults to "{DEFAULT_SERVER_ADDRESS}".')
@click.option('--config', '-c', 'config_paths',
              metavar='CONFIG', default=None, multiple=True,
              help='Configuration YAML or JSON file. '
                   ' If multiple are passed,'
                   ' they will be merged in order.'
                   ' Can be a local filesystem path or an absolute URL.')
@click.option('--base-dir', 'base_dir',
              metavar='BASE_DIR', default=None,
              help='Directory used to resolve relative paths'
                   ' in CONFIG files. Defaults to the parent directory'
                   ' of (last) CONFIG file.'
                   ' Can be a local filesystem path or an absolute URL.')
@click.option('--prefix', 'url_prefix',
              metavar='URL_PREFIX', default=None,
              help='Prefix path to be used for all endpoint URLs.'
                   ' May include template variables, e.g., "api/{version}".'
                   ' Can be an absolute URL or a relative URL path.')
@click.option('--revprefix', 'reverse_url_prefix',
              metavar='REVERSE_URL_PREFIX', default=None,
              help='Prefix path to be used for reverse endpoint URLs'
                   ' that may be reported by server responses.'
                   ' May include template variables, e.g., "/proxy/{port}".'
                   ' Defaults to value of URL_PREFIX.'
                   ' Can be an absolute URL or relative URL path.')
@click.option('--traceperf', 'trace_perf', is_flag=True, default=None,
              help='Whether to output extra performance logs.')
@click.option('--update-after', 'update_after',
              metavar='TIME', type=float, default=None,
              help='Check for server configuration updates every'
                   ' TIME seconds.')
@click.option('--stop-after', 'stop_after',
              metavar='TIME', type=float, default=None,
              help='Unconditionally stop service after TIME seconds.')
@click.option('--show', 'asset_to_show',
              type=click.Choice(assets_to_show_choices),
              metavar='ASSET', nargs=1,
              help=f"Show ASSET and exit. Possible values for ASSET are"
                   f" {', '.join(map(repr, assets_to_show))} optionally"
                   f" suffixed by '.yaml' or '.json'.")
@click.option('--open-viewer', 'open_viewer',
              is_flag=True,
              help=f"After starting the server,"
                   f" open xcube Viewer in a browser tab.")
@cli_option_quiet
@cli_option_verbosity
@click.argument('paths', metavar='[PATHS...]', nargs=-1)
def serve(framework_name: str,
          port: int,
          address: str,
          config_paths: List[str],
          base_dir: Optional[str],
          url_prefix: Optional[str],
          reverse_url_prefix: Optional[str],
          trace_perf: Optional[bool],
          update_after: Optional[float],
          stop_after: Optional[float],
          asset_to_show: str,
          open_viewer: bool,
          quiet: bool,
          verbosity: int,
          paths: List[str]):
    """
    Run the xcube Server for the given configuration and/or the given
    raster dataset paths given by PATHS.

    Each of the PATHS arguments can point to a raster dataset such as a Zarr
    directory (*.zarr), an xcube multi-level Zarr dataset (*.levels),
    a NetCDF file (*.nc), a GeoTIFF/COG file (*.tiff).

    If one of PATHS is a directory that is not a dataset itself,
    it is scanned for readable raster datasets.

    The --show ASSET option can be used to inspect the current configuration
    of the server. ASSET is one of:

    \b
    apis            outputs the list of APIs provided by the server
    endpoints       outputs the list of all endpoints provided by the server
    openapi         outputs the OpenAPI document representing this server
    config          outputs the effective server configuration
    configschema    outputs the JSON Schema for the server configuration

    The ASSET may be suffixed by ".yaml" or ".json"
    forcing the respective output format. The default format is YAML.

    Note, if --show  is provided, the ASSET will be shown and the program
    will exit immediately.
    """
    from xcube.server.framework import get_framework_class
    from xcube.server.helpers import ConfigChangeObserver
    from xcube.server.server import Server
    from xcube.server.config import normalize_base_dir
    from xcube.util.config import load_configs

    configure_cli_output(quiet=quiet, verbosity=verbosity)

    config = load_configs(*config_paths,
                          exception_type=click.ClickException) \
        if config_paths else {}

    if port is not None:
        config["port"] = port
    port = config.get("port")
    if port is None:
        port = DEFAULT_SERVER_PORT
        config["port"] = port

    if address is not None:
        config["address"] = address
    address = config.get("address")
    if address is None:
        address = DEFAULT_SERVER_ADDRESS
        config["address"] = address

    if base_dir is not None:
        # Use base_dir CLI option
        pass
    elif "base_dir" in config:
        # Use base_dir from configuration
        base_dir = config["base_dir"]
    elif config_paths:
        # Use base_dir derived from last config file's parent
        config_path = config_paths[-1].replace('\\', '/')
        base_dir = '/'.join(config_path.split('/')[:-1])
    else:
        # base_dir is current working directory
        base_dir = str(Path.cwd())
    base_dir = normalize_base_dir(base_dir)
    config["base_dir"] = base_dir

    if url_prefix is not None:
        config["url_prefix"] = url_prefix

    if reverse_url_prefix is not None:
        config["reverse_url_prefix"] = reverse_url_prefix

    if trace_perf is not None:
        config["trace_perf"] = trace_perf

    if paths:
        data_stores: Dict[Path, Dict[str, Any]] = dict()
        for path in paths:
            if Path(path).exists():
                path = Path(path)
                if path.is_dir():
                    if path.suffix == '.zarr' or path.suffix == '.levels':
                        _add_path_to_data_stores(data_stores, path)
                    else:
                        _add_dir_to_data_stores(data_stores, path)
                else:
                    _add_path_to_data_stores(data_stores, path)
            else:
                # TODO (forman): Support also "s3://" prefixes
                raise click.ClickException(
                    f"File or directory not found: {path}"
                )
        config["DataStores"] = list(data_stores.values())

    framework = get_framework_class(framework_name)()
    server = Server(framework, config)

    if asset_to_show:
        return show_asset(server, asset_to_show)

    if update_after is not None:
        change_observer = ConfigChangeObserver(server,
                                               config_paths,
                                               update_after)
        server.call_later(update_after, change_observer.check)

    if stop_after is not None:
        server.call_later(stop_after, server.stop)

    if open_viewer:
        def open_xcube_viewer():
            import webbrowser
            server_url = f'http://localhost:{port}'
            webbrowser.open_new_tab(f'{server_url}/viewer/'
                                    f'?serverUrl={server_url}'
                                    f'&serverId=local'
                                    f'&serverName=Local')

        server.call_later(2.5, open_xcube_viewer)

    server.start()


def _add_dir_to_data_stores(data_stores: Dict[Path, Dict[str, Any]],
                            dir_path: Path) -> Dict[str, Any]:
    root = dir_path.resolve().absolute()
    if root in data_stores:
        data_store = data_stores[root]
    else:
        data_store = data_stores[root] = dict(
            Identifier=str(root.name),
            StoreId="file",
            StoreParams=dict(root=str(root))
        )
    return data_store


def _add_path_to_data_stores(data_stores: Dict[Path, Dict[str, Any]],
                             path: Path):
    path = path.resolve().absolute()
    data_store = _add_dir_to_data_stores(data_stores, path.parent)
    if "Datasets" in data_store:
        datasets = data_store["Datasets"]
    else:
        datasets = data_store["Datasets"] = []
    datasets.append(dict(
        Identifier=str(path.name),
        Path=str(path.name)
    ))


def show_asset(server, asset_to_show: str):
    def to_yaml(obj):
        import yaml
        yaml.safe_dump(obj, sys.stdout, indent=2)

    def to_json(obj):
        import json
        json.dump(obj, sys.stdout, indent=2)

    available_formats = {
        'yaml': to_yaml,
        'json': to_json
    }

    format_name = 'yaml'
    splits = asset_to_show.rsplit('.', maxsplit=1)
    if len(splits) == 2:
        asset_to_show, format_name = splits

    if format_name.lower() not in available_formats:
        raise click.ClickException(
            f'Invalid format {format_name}.'
            f' Must be one of {", ".join(available_formats.keys())}.'
        )
    output_fn = available_formats[format_name.lower()]

    def show_apis():
        output_fn([dict(name=api.name,
                        version=api.version,
                        description=api.description)
                   for api in server.apis])

    def show_endpoints():
        output_fn([route.path
                   for api in server.apis
                   for route in (api.routes + api.static_routes)])

    def show_open_api():
        output_fn(server.ctx.get_open_api_doc())

    def show_config():
        output_fn(server.ctx.config.defrost())

    def show_config_schema():
        output_fn(server.config_schema.to_dict())

    available_commands = {
        "apis": show_apis,
        "endpoints": show_endpoints,
        "openapi": show_open_api,
        "config": show_config,
        "configschema": show_config_schema,
    }

    command_fn = available_commands.get(asset_to_show)
    if command_fn is None:
        raise click.ClickException(
            f'Invalid command {asset_to_show}. '
            f'Possible commands are '
            f'{", ".join(" ".join(k) for k in available_commands.keys())}.'
        )
    return command_fn()



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
import datetime
import hashlib
import time
from typing import Dict, Any, List, Mapping, Optional

from xcube.util.assertions import assert_instance

_CONTENT_LENGTH_DUMMY = -1
_LAST_MODIFIED_DUMMY = str(datetime.datetime.utcfromtimestamp(time.time()))


def list_s3_bucket_v2(object_storage: Mapping[str, bytes],
                      name: Optional[str] = None,
                      delimiter: Optional[str] = None,
                      prefix: Optional[str] = None,
                      max_keys: Optional[str] = None,
                      start_after: Optional[str] = None,
                      continuation_token: Optional[str] = None,
                      storage_class: Optional[str] = None,
                      last_modified: Optional[str] = None) -> Dict:
    """Implements AWS GET Bucket (List Objects) Version 2
    (https://docs.aws.amazon.com/AmazonS3/latest/API/v2-RESTBucketGET.html)
    for the local filesystem.

    :param object_storage: mapping from path-like keys to byte-objects.
    :param name: The bucket name.
    :param delimiter: A delimiter is a character you use to group keys.
        If you specify a prefix, all of the keys that contain the same string
        between the prefix and the first occurrence of the delimiter after
        the prefix are grouped under a single result element called
        CommonPrefixes.
        If you don't specify the prefix parameter, the substring starts
        at the beginning of the key.
        The keys that are grouped under the CommonPrefixes result element
        are not returned elsewhere in the response.
        Refer to AWS docs for details. No default.
    :param prefix: Limits the response to keys that begin with the
        specified prefix.
        You can use prefixes to separate a bucket into different
        groupings of keys. (You can think of using prefix to make groups
        in the same way you'd use a folder in a file system.)
        Refer to AWS docs for details. No default.
    :param max_keys: Sets the maximum number of keys returned
        in the response body.
        If you want to retrieve fewer than the default 1000 keys,
        you can add this to your request.
        The response might contain fewer keys, but it never contains more.
        If there are additional keys that satisfy the search criteria,
        but these keys were not returned because max-keys was exceeded,
        the response contains <IsTruncated>true</IsTruncated>.
        To return the additional keys, see NextContinuationToken.
        Refer to AWS docs for details. Defaults to 1000.
    :param start_after: If you want the API to return key names after
        a specific object key in your key space, you can add this parameter.
        Amazon S3 lists objects in UTF-8 character encoding
        in lexicographical order. Refer to AWS docs for details.
        No default.
    :param continuation_token: When the response to this API call is
        truncated (that is, the IsTruncated response element value is true),
        the response also includes the NextContinuationToken element.
        To list the next set of objects,
        you can use the NextContinuationToken element in the next request
        as the continuation-token.
        Refer to AWS docs for details. No default.
    :param storage_class: Refer to AWS docs for details.
        Defaults to "STANDARD".
    :param last_modified: For testing only: always use this value
        for the "LastModified" entry of results
    :return: A dictionary that represents the contents of a
        "ListBucketResult". Refer to AWS docs for details.
    """
    assert_instance(object_storage,
                    collections.abc.Mapping,
                    name="object_storage")
    assert_instance(name, str, name="name")

    max_keys = max_keys or 1000
    start_after = None if continuation_token else start_after
    storage_class = storage_class or 'STANDARD'

    contents_list = []
    next_continuation_token = None
    continuation_token_seen = continuation_token is None
    start_key_seen = start_after is None
    common_prefixes_list = []
    common_prefixes_set = set()

    token = 0

    for key in sorted(object_storage.keys()):

        token += 1

        if len(contents_list) == max_keys:
            next_continuation_token = token
            break

        if key == start_after:
            start_key_seen = True
            continue

        if token == continuation_token:
            continuation_token_seen = True

        if not continuation_token_seen or not start_key_seen:
            continue

        if prefix and not key.startswith(prefix):
            continue

        if delimiter:
            index = key.find(delimiter, len(prefix) if prefix else 0)
            if index >= 0:
                key = key[:index + len(delimiter)]
                if key not in common_prefixes_set:
                    common_prefixes_set.add(key)
                    common_prefixes_list.append(key)
                continue

        item = dict(
            Key=key,
            Size=_CONTENT_LENGTH_DUMMY,
            LastModified=last_modified or _LAST_MODIFIED_DUMMY,
            ETag=_str_to_e_tag(key),
            StorageClass=storage_class
        )
        contents_list.append(item)

    list_bucket_result: Dict[str, Any] = dict(
        Name=name,
        Prefix=prefix,
        StartAfter=start_after,
        MaxKeys=max_keys,
        Delimiter=delimiter,
        IsTruncated=next_continuation_token is not None,
        ContinuationToken=continuation_token
    )
    if next_continuation_token is not None:
        list_bucket_result.update(
            NextContinuationToken=next_continuation_token
        )
    if contents_list:
        list_bucket_result.update(
            Contents=contents_list
        )
    if common_prefixes_list:
        list_bucket_result.update(
            CommonPrefixes=[dict(Prefix=prefix)
                            for prefix in common_prefixes_list]
        )
    return list_bucket_result


def list_s3_bucket_v1(object_storage: Mapping[str, bytes],
                      name: str = None,
                      delimiter: str = None,
                      prefix: str = None,
                      max_keys: int = None,
                      marker: str = None,
                      storage_class: str = None,
                      last_modified: str = None) -> Dict:
    """Implements AWS GET Bucket (List Objects) Version 1
    (https://docs.aws.amazon.com/AmazonS3/latest/API/RESTBucketGET.html)
    for the local filesystem.

    :param object_storage: mapping from path-like keys to byte-objects.
    :param name: The bucket name, defaults to "s3bucket"
    :param delimiter: A delimiter is a character you use to group keys.
        If you specify a prefix, all of the keys that contain the same string
        between the prefix and the first occurrence of the delimiter after
        the prefix are grouped under a single result element called 
        CommonPrefixes.
        If you don't specify the prefix parameter, the substring starts 
        at the beginning of the key.
        The keys that are grouped under the CommonPrefixes result 
        element are not returned elsewhere in the response.
        Refer to AWS docs for details. No default.
    :param prefix: Limits the response to keys that begin with the 
        specified prefix.
        You can use prefixes to separate a bucket into different 
        groupings of keys. (You can think of using prefix to make groups
        in the same way you'd use a folder in a file system.)
        Refer to AWS docs for details. No default.
    :param max_keys: Sets the maximum number of keys returned in the 
        response body. If you want to retrieve fewer than the default 
        1000 keys, you can add this to your request.
        The response might contain fewer keys, but it never contains more.
        If there are additional keys that satisfy the search criteria, 
        but these keys were not returned because max-keys was exceeded, 
        the response contains <IsTruncated>true</IsTruncated>.
        To return the additional keys, see NextMarker.
        Refer to AWS docs for details.
    :param marker: Indicates the object key to start with when listing 
        objects in a bucket. All objects are listed in the dictionary order.
        Refer to AWS docs for details. No default.
    :param storage_class: Refer to AWS docs for details. 
        Defaults to "STANDARD".
    :param last_modified: For testing only: always use this value 
        for the "LastModified" entry of results
    :return: A dictionary that represents the contents of
        a "ListBucketResult". Refer to AWS docs for details.
    """
    assert_instance(object_storage,
                    collections.abc.Mapping,
                    name="object_storage")
    assert_instance(name, str, name="name")

    max_keys = max_keys or 1000
    storage_class = storage_class or 'STANDARD'

    contents_list = []
    is_truncated = False
    next_marker = None
    marker_seen = marker is None
    common_prefixes_list = []
    common_prefixes_set = set()

    for key in sorted(object_storage.keys()):

        if len(contents_list) == max_keys:
            is_truncated = True
            next_marker = key
            break

        if key == marker:
            marker_seen = True

        if not marker_seen:
            continue

        if prefix and not key.startswith(prefix):
            continue

        if delimiter:
            index = key.find(delimiter, len(prefix) if prefix else 0)
            if index >= 0:
                key = key[:index + len(delimiter)]
                if key not in common_prefixes_set:
                    common_prefixes_set.add(key)
                    common_prefixes_list.append(key)
                continue

        item = dict(
            Key=key,
            Size=_CONTENT_LENGTH_DUMMY,
            LastModified=last_modified or _LAST_MODIFIED_DUMMY,
            ETag=_str_to_e_tag(key),
            StorageClass=storage_class
        )
        contents_list.append(item)

    list_bucket_result: Dict[str, Any] = dict(
        Name=name,
        Prefix=prefix,
        Marker=marker,
        MaxKeys=max_keys,
        Delimiter=delimiter,
        IsTruncated=is_truncated
    )
    if is_truncated:
        list_bucket_result.update(NextMarker=next_marker)
    if contents_list:
        list_bucket_result.update(Contents=contents_list)
    if common_prefixes_list:
        list_bucket_result.update(
            CommonPrefixes=[dict(Prefix=prefix)
                            for prefix in common_prefixes_list]
        )
    return list_bucket_result


def list_bucket_result_to_xml(list_bucket_result):
    return dict_to_xml(
        'ListBucketResult',
        list_bucket_result,
        root_element_attrs=dict(
            xmlns="http://s3.amazonaws.com/doc/2006-03-01/"
        )
    )


def dict_to_xml(root_element_name: str,
                content_dict: Dict,
                root_element_attrs: Dict = None) -> str:
    lines = []
    _value_to_xml(lines, root_element_name, content_dict,
                  element_attrs=root_element_attrs)
    return '<?xml version="1.0" encoding="UTF-8"?>\n' + '\n'.join(lines)


def _value_to_xml(lines: List[str],
                  element_name: str,
                  element_value: Any,
                  element_attrs: Dict = None,
                  indent: int = 0):
    attrs = ''
    if element_attrs:
        for attr_name, attr_value in element_attrs.items():
            attrs += f' {attr_name}=\"{_value_to_text(attr_value)}\"'
    if element_value is None:
        lines.append(f'{indent * "  "}<{element_name}{attrs}/>')
    elif isinstance(element_value, dict):
        lines.append(f'{indent * "  "}<{element_name}{attrs}>')
        for sub_element_name, sub_element_value in element_value.items():
            _value_to_xml(lines,
                          sub_element_name,
                          sub_element_value,
                          indent=indent + 1)
        lines.append(f'{indent * "  "}</{element_name}>')
    elif isinstance(element_value, list):
        for item in element_value:
            _value_to_xml(lines,
                          element_name,
                          item,
                          indent=indent)
    else:
        lines.append(f'{indent * "  "}'
                     f'<{element_name}{attrs}>'
                     f'{_value_to_text(element_value)}'
                     f'</{element_name}>')


def _value_to_text(value) -> str:
    if isinstance(value, bool):
        return 'true' if value else 'false'
    return str(value)


def _str_to_e_tag(s) -> str:
    return '"' + hashlib.md5(bytes(s, encoding='utf-8')).hexdigest() + '"'



import itertools
import os
import re
import uuid
import warnings
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, \
    Sequence, Tuple, Union

import dask.array as da
import dask.array.core as dac
import distributed
import numpy as np

IntTuple = Tuple[int, ...]
SliceTuple = Tuple[slice, ...]
IntIterable = Iterable[int]
IntTupleIterable = Iterable[IntTuple]
SliceTupleIterable = Iterable[SliceTuple]

_CLUSTER_TAGS_ENV_VAR_NAME = 'XCUBE_DASK_CLUSTER_TAGS'
_CLUSTER_ACCOUNT_ENV_VAR_NAME = 'XCUBE_DASK_CLUSTER_ACCOUNT'


def compute_array_from_func(func: Callable[..., np.ndarray],
                            shape: IntTuple,
                            chunks: IntTuple,
                            dtype: Any,
                            name: str = None,
                            ctx_arg_names: Sequence[str] = None,
                            args: Sequence[Any] = None,
                            kwargs: Mapping[str, Any] = None) -> da.Array:
    """
    Compute a dask array using the provided user function *func*, *shape*, and chunking *chunks*.

    The user function is expected to output the array's data blocks using arguments specified by
    *ctx_arg_names*, *args*, and *kwargs* and is expected to return a numpy array.

    You can request array and current block context information by specifying the optional
    *ctx_arg_names* keyword argument that is a sequence of names of special arguments passed to *user_func*.
    The following are available:

    * ``shape``: The array's shape. A tuple of ints.
    * ``chunks``: The array's chunks. A tuple of tuple of ints.
    * ``dtype``: The array's numpy data type.
    * ``name``: The array's name. A string or ``None``.
    * ``block_id``: The block's unique ID. An integer number ranging from zero to number of blocks minus one.
    * ``block_index``: The block's index as a tuple of ints.
    * ``block_shape``: The block's shape as a tuple of ints.
    * ``block_slices``: The block's shape as a tuple of int pair tuples.

    :param func: User function that is called for each block of the array using arguments specified by
        *ctx_arg_names*, *args*, and *kwargs*. It must return a numpy array of shape "block_shape" and type
        *dtype*.
    :param shape: The array's shape. A tuple of sizes for each array dimension.
    :param chunks: The array's chunking. A tuple of chunk sizes for each array dimension.
        Must be of same length as *shape*.
    :param dtype: The array's numpy data type.
    :param name: The array's name.
    :param ctx_arg_names: Sequence names of arguments that are passed
        before *args* and *kwargs* to the user function.
    :param args: Arguments passed to the user function.
    :param kwargs: Keyword-arguments passed to the user function.
    :return: A chunked dask array.
    """
    ctx_arg_names = ctx_arg_names or []
    args = args or []
    kwargs = kwargs or {}

    chunk_sizes = tuple(get_chunk_sizes(shape, chunks))
    chunk_counts = tuple(get_chunk_counts(shape, chunks))
    block_indexes, block_shapes, block_slices = get_block_iterators(chunk_sizes)

    ctx_values = dict(
        shape=tuple(shape),
        chunks=chunk_sizes,
        dtype=dtype,
        name=name,
    )

    blocks = _NestedList(shape=chunk_counts)
    block_id = 0
    for chunk_index, chunk_shape, block_slices in zip(block_indexes, block_shapes, block_slices):
        ctx_values.update(
            block_id=block_id,
            block_index=tuple(chunk_index),
            block_shape=tuple(chunk_shape),
            block_slices=tuple((chunk_slice.start, chunk_slice.stop) for chunk_slice in block_slices),
        )
        ctx_args = [ctx_values[ctx_arg_name] for ctx_arg_name in ctx_arg_names]
        block_id += 1

        # We use our own name here, because dac.from_func() tokenizes args which for some reason takes forever
        block = dac.from_func(func,
                              shape=chunk_shape,
                              dtype=dtype,
                              name=f'rectify_{name}-{uuid.uuid4()}',
                              args=(*ctx_args, *args),
                              kwargs=kwargs)

        blocks[chunk_index] = block

    return da.block(blocks.data)


def get_block_iterators(chunk_sizes: IntTupleIterable) -> \
        Tuple[IntTupleIterable, IntTupleIterable, SliceTupleIterable]:
    chunk_sizes = tuple(chunk_sizes)
    chunk_slices_tuples = get_chunk_slice_tuples(chunk_sizes)
    chunk_ranges = get_chunk_ranges(chunk_sizes)
    block_indexes = itertools.product(*chunk_ranges)
    block_shapes = itertools.product(*chunk_sizes)
    block_slices = itertools.product(*chunk_slices_tuples)
    return block_indexes, block_shapes, block_slices


def get_chunk_sizes(shape: IntTuple, chunks: IntTuple) -> IntTupleIterable:
    for s, c in zip(shape, chunks):
        n = s // c
        if n * c < s:
            yield (c,) * n + (s % c,)
        else:
            yield (c,) * n


def get_chunk_counts(shape: IntTuple, chunks: IntTuple) -> Iterable[int]:
    for s, c in zip(shape, chunks):
        yield (s + c - 1) // c


def get_chunk_ranges(chunk_size_tuples: IntTupleIterable) -> Iterable[range]:
    return (range(len(chunk_size_tuple)) for chunk_size_tuple in chunk_size_tuples)


def get_chunk_slice_tuples(chunk_size_tuples: IntTupleIterable) -> SliceTupleIterable:
    return (tuple(get_chunk_slices(chunk_size_tuple)) for chunk_size_tuple in chunk_size_tuples)


def get_chunk_slices(chunk_sizes: Sequence[int]) -> Iterable[slice]:
    stop = 0
    for i in range(len(chunk_sizes)):
        start = stop
        stop = start + chunk_sizes[i]
        yield slice(start, stop)


def new_cluster(
    provider: str = 'coiled',
    name: Optional[str] = None,
    software: Optional[str] = None,
    n_workers: int = 4,
    resource_tags: Optional[Dict[str, str]] = None,
    account: str = None,
    region: str = 'eu-central-1',
    **kwargs,
) -> distributed.deploy.Cluster:
    """Create a new Dask cluster.

    Cloud resource tags can be specified in an environment variable
    XCUBE_DASK_CLUSTER_TAGS in the format
    ``tag_1=value_1:tag_2=value_2:...:tag_n=value_n``. In case of
    conflicts, tags specified in ``resource_tags`` will override tags
    specified by the environment variable.

    The cluster provider account name can be specified in an environment
    variable ``XCUBE_DASK_CLUSTER_ACCOUNT``. If the ``account`` argument is
    given to ``new_cluster``, it will override the value from the environment
    variable.

    :param provider: identifier of the provider to use. Currently, only
        'coiled' is supported.
    :param name: name to use as an identifier for the cluster
    :param software: identifier for the software environment to be used.
    :param n_workers: number of workers in the cluster
    :param resource_tags: tags to apply to the cloud resources forming the
        cluster
    :param account: cluster provider account name
    :param **kwargs: further named arguments will be passed on to the
        cluster creation function
    :param region: default region where workers of the cluster will be deployed set to eu-central-1
    """

    if resource_tags is None:
        resource_tags = {}
    if _CLUSTER_ACCOUNT_ENV_VAR_NAME in os.environ:
        account_from_env_var = os.environ[_CLUSTER_ACCOUNT_ENV_VAR_NAME]
    else:
        account_from_env_var = None
        warnings.warn(f'Environment variable {_CLUSTER_ACCOUNT_ENV_VAR_NAME}'
                      f' not set; cluster account name may be incorrect.')

    cluster_account = (
        account if account is not None else
        account_from_env_var if account_from_env_var is not None else
        'bc'
    )

    if provider == 'coiled':
        try:
            import coiled
        except ImportError as e:
            raise ImportError(f"provider 'coiled' requires package"
                              f"'coiled' to be installed") from e
        if software is None and 'JUPYTER_IMAGE' in os.environ:
            # If the JUPYTER_IMAGE environment variable is set, we're
            # presumably in a Z2JH deployment and can base a
            # Coiled environment on the same image.
            # First we construct an identifier from the user image specifier.
            current_image = os.environ['JUPYTER_IMAGE']
            software = re.sub(
                '[:.]',
                '-',
                re.search(r'/([^/]+)$', current_image).group(1),
            )
            # If the referenced software environment doesn't exist yet as a
            # Coiled environment, create it from the currently used image.
            available_environments = \
                coiled.list_software_environments(account=account).keys()
            if software not in available_environments:
                coiled.create_software_environment(
                    name=software,
                    container=current_image
                )

        # If software is (still) None, Coiled will try to mirror the current
        # environment automagically.
        coiled_params = dict(
            n_workers=n_workers,
            environ=None,
            tags=_collate_cluster_resource_tags(resource_tags),
            account=cluster_account,
            name=name,
            software=software,
            use_best_zone=True,
            compute_purchase_option='spot_with_fallback',
            shutdown_on_close=True,
            region=region,
        )
        coiled_params.update(kwargs)

        return coiled.Cluster(**coiled_params)

    raise NotImplementedError(f'Unknown provider {provider!r}')


def _collate_cluster_resource_tags(extra_tags: Dict[str, str]) \
        -> Dict[str, str]:
    fallback_tags = {
        'cost-center': 'unknown',
        'environment': 'dev',
        'creator': 'auto',
        'purpose': 'xcube dask cluster',
        'user': (os.environ.get('JUPYTERHUB_USER')  # JupyterHub
                 or os.environ.get('USER')          # Unixes
                 or os.environ.get('USERNAME')      # Windows
                 or os.getlogin()
                 or '')
    }
    if _CLUSTER_TAGS_ENV_VAR_NAME in os.environ:
        kvps = os.environ[_CLUSTER_TAGS_ENV_VAR_NAME].split(':')
        env_var_tags = {
            (parts := kvp.split('=', maxsplit=1))[0]: parts[1] for kvp in kvps
        }
    else:
        warnings.warn(f'Environment variable {_CLUSTER_TAGS_ENV_VAR_NAME}'
                      f' not set; cluster resource tags may be missing.')
        env_var_tags = {}
    return fallback_tags | env_var_tags | extra_tags


class _NestedList:
    """
    Utility class whose instances are used as input to dask.block().
    """

    def __init__(self, shape: Sequence[int], fill_value: Any = None):
        self._shape = tuple(shape)
        self._data = self._new_data(shape, len(shape), fill_value, 0)

    @classmethod
    def _new_data(cls, shape: Sequence[int], ndim: int, fill_value: Any, dim: int) -> Union[List[List], List[Any]]:
        return [cls._new_data(shape, ndim, fill_value, dim + 1) if dim < ndim - 1 else fill_value
                for _ in range(shape[dim])]

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @property
    def data(self) -> Union[List[List], List[Any]]:
        return self._data

    def __len__(self) -> int:
        return len(self._data)

    def __setitem__(self, index: Union[int, slice, tuple], value: Any):
        data = self._data
        if isinstance(index, tuple):
            n = len(index)
            for i in range(n - 1):
                data = data[index[i]]
            data[index[n - 1]] = value
        else:
            data[index] = value

    def __getitem__(self, index: Union[int, slice, tuple]) -> Any:
        data = self._data
        if isinstance(index, tuple):
            n = len(index)
            for i in range(n - 1):
                data = data[index[i]]
            return data[index[n - 1]]
        else:
            return data[index]



#  The MIT License (MIT)
#  Copyright (c) 2022 by the xcube development team and contributors
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

# noinspection PyUnresolvedReferences
from .api import api



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Mapping, Optional, Sequence, Tuple, Union

import dask.array as da
import numba as nb
import numpy as np
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.select import select_spatial_subset
from xcube.util.dask import compute_array_from_func
from .cf import maybe_encode_grid_mapping


def rectify_dataset(source_ds: xr.Dataset,
                    *,
                    var_names: Union[str, Sequence[str]] = None,
                    source_gm: GridMapping = None,
                    xy_var_names: Tuple[str, str] = None,
                    target_gm: GridMapping = None,
                    encode_cf: bool = True,
                    gm_name: Optional[str] = None,
                    tile_size: Union[int, Tuple[int, int]] = None,
                    is_j_axis_up: bool = None,
                    output_ij_names: Tuple[str, str] = None,
                    compute_subset: bool = True,
                    uv_delta: float = 1e-3) -> Optional[xr.Dataset]:
    """
    Reproject dataset *source_ds* using its per-pixel
    x,y coordinates or the given *source_gm*.

    The function expects *source_ds* or the given
    *source_gm* to have either one- or two-dimensional
    coordinate variables that provide spatial x,y coordinates
    for every data variable with the same spatial dimensions.

    For example, a dataset may comprise variables with
    spatial dimensions ``var(..., y_dim, x_dim)``,
    then one the function expects coordinates to be provided
    in two forms:

    1. One-dimensional ``x_var(x_dim)``
       and ``y_var(y_dim)`` (coordinate) variables.
    2. Two-dimensional ``x_var(y_dim, x_dim)``
       and ``y_var(y_dim, x_dim)`` (coordinate) variables.

    If *target_gm* is given and it defines a tile size
    or *tile_size* is given, and the number of tiles is
    greater than one in the output's x- or y-direction, then the
    returned dataset will be composed of lazy, chunked dask
    arrays. Otherwise, the returned dataset will be composed
    of ordinary numpy arrays.

    :param source_ds: Source dataset grid mapping.
    :param var_names: Optional variable name or sequence of
        variable names.
    :param source_gm: Source dataset grid mapping.
    :param xy_var_names: Optional tuple of the x- and y-coordinate
        variables in *source_ds*. Ignored if *source_gm* is given.
    :param target_gm: Optional target geometry. If not given,
        output geometry will be computed to spatially fit *dataset*
        and to retain its spatial resolution.
    :param encode_cf: Whether to encode the target grid mapping
        into the resampled dataset in a CF-compliant way.
        Defaults to ``True``.
    :param gm_name: Name for the grid mapping variable.
        Defaults to "crs". Used only if *encode_cf* is ``True``.
    :param tile_size: Optional tile size for the output.
    :param is_j_axis_up: Whether y coordinates are increasing with
        positive image j axis.
    :param output_ij_names: If given, a tuple of variable names in
        which to store the computed source pixel coordinates in
        the returned output.
    :param compute_subset: Whether to compute a spatial subset
        from *dataset* using *output_geom*. If set, the function
        may return ``None`` in case there is no overlap.
    :param uv_delta: A normalized value that is used to determine
        whether x,y coordinates in the output are contained
        in the triangles defined by the input x,y coordinates.
        The higher this value, the more inaccurate the rectification
        will be.
    :return: a reprojected dataset, or None if the requested output
        does not intersect with *dataset*.
    """
    if source_gm is None:
        source_gm = GridMapping.from_dataset(source_ds)

    src_attrs = dict(source_ds.attrs)

    if target_gm is None:
        target_gm = source_gm.to_regular(tile_size=tile_size)
    elif compute_subset:
        source_ds_subset = select_spatial_subset(
            source_ds,
            xy_bbox=target_gm.xy_bbox,
            ij_border=1,
            xy_border=0.5 * (target_gm.x_res + target_gm.y_res),
            grid_mapping=source_gm
        )
        if source_ds_subset is None:
            return None
        if source_ds_subset is not source_ds:
            # TODO: GridMapping.from_dataset() may be expensive.
            #   Find a more effective way.
            source_gm = GridMapping.from_dataset(source_ds_subset)
            source_ds = source_ds_subset

    # if src_geo_coding.xy_var_names != output_geom.xy_var_names:
    #     output_geom = output_geom.derive(
    #           xy_var_names=src_geo_coding.xy_var_names
    #     )
    # if src_geo_coding.xy_dim_names != output_geom.xy_dim_names:
    #     output_geom = output_geom.derive(
    #           xy_dim_names=src_geo_coding.xy_dim_names
    #     )

    if tile_size is not None or is_j_axis_up is not None:
        target_gm = target_gm.derive(tile_size=tile_size,
                                     is_j_axis_up=is_j_axis_up)

    src_vars = _select_variables(source_ds, source_gm, var_names)

    if target_gm.is_tiled:
        compute_dst_src_ij_images = _compute_ij_images_xarray_dask
        compute_dst_var_image = _compute_var_image_xarray_dask
    else:
        compute_dst_src_ij_images = _compute_ij_images_xarray_numpy
        compute_dst_var_image = _compute_var_image_xarray_numpy

    dst_src_ij_array = compute_dst_src_ij_images(source_gm,
                                                 target_gm,
                                                 uv_delta)

    dst_x_dim, dst_y_dim = target_gm.xy_dim_names
    dst_dims = dst_y_dim, dst_x_dim
    dst_ds_coords = target_gm.to_coords()
    dst_vars = dict()
    for src_var_name, src_var in src_vars.items():
        dst_var_dims = src_var.dims[0:-2] + dst_dims
        dst_var_coords = {d: src_var.coords[d]
                          for d in dst_var_dims if d in src_var.coords}
        dst_var_coords.update({d: dst_ds_coords[d]
                               for d in dst_var_dims if d in dst_ds_coords})
        dst_var_array = compute_dst_var_image(src_var,
                                              dst_src_ij_array,
                                              fill_value=np.nan)
        dst_var = xr.DataArray(dst_var_array,
                               dims=dst_var_dims,
                               coords=dst_var_coords,
                               attrs=src_var.attrs)
        dst_vars[src_var_name] = dst_var

    if output_ij_names:
        output_i_name, output_j_name = output_ij_names
        dst_ij_coords = {d: dst_ds_coords[d]
                         for d in dst_dims if d in dst_ds_coords}
        dst_vars[output_i_name] = xr.DataArray(dst_src_ij_array[0],
                                               dims=dst_dims,
                                               coords=dst_ij_coords)
        dst_vars[output_j_name] = xr.DataArray(dst_src_ij_array[1],
                                               dims=dst_dims,
                                               coords=dst_ij_coords)

    return maybe_encode_grid_mapping(encode_cf,
                                     xr.Dataset(dst_vars,
                                                coords=dst_ds_coords,
                                                attrs=src_attrs),
                                     target_gm,
                                     gm_name)


def _select_variables(
        source_ds: xr.Dataset,
        source_gm: GridMapping,
        var_names: Union[None, str, Sequence[str]]
) -> Mapping[str, xr.DataArray]:
    """
    Select variables from *dataset*.

    :param source_ds: Source dataset.
    :param source_gm: Optional dataset geo-coding.
    :param var_names: Optional variable name
        or sequence of variable names.
    :return: The selected variables as a variable name
        to ``xr.DataArray`` mapping
    """
    spatial_var_names = source_gm.xy_var_names
    spatial_shape = tuple(reversed(source_gm.size))
    spatial_dims = tuple(reversed(source_gm.xy_dim_names))
    if var_names is None:
        var_names = [var_name
                     for var_name, var in source_ds.data_vars.items()
                     if var_name not in spatial_var_names
                     and _is_2d_spatial_var(var, spatial_shape, spatial_dims)]
    elif isinstance(var_names, str):
        var_names = (var_names,)
    elif len(var_names) == 0:
        raise ValueError(f'empty var_names')
    src_vars = {}
    for var_name in var_names:
        src_var = source_ds[var_name]
        if not _is_2d_spatial_var(src_var, spatial_shape, spatial_dims):
            raise ValueError(
                f"cannot rectify variable {var_name!r}"
                f" as its shape or dimensions "
                f"do not match those of {spatial_var_names[0]!r}"
                f" and {spatial_var_names[1]!r}"
            )
        src_vars[var_name] = src_var
    return src_vars


def _is_2d_spatial_var(var: xr.DataArray, shape, dims) -> bool:
    return var.ndim >= 2 and var.shape[-2:] == shape and var.dims[-2:] == dims


def _compute_ij_images_xarray_numpy(src_geo_coding: GridMapping,
                                    output_geom: GridMapping,
                                    uv_delta: float) -> np.ndarray:
    """
    Compute numpy.ndarray destination image with source
    pixel i,j coords from xarray.DataArray x,y sources.
    """
    dst_width = output_geom.width
    dst_height = output_geom.height
    dst_shape = 2, dst_height, dst_width
    dst_src_ij_images = np.full(dst_shape, np.nan, dtype=np.float64)
    dst_x_offset = output_geom.x_min
    dst_y_offset = output_geom.y_min \
        if output_geom.is_j_axis_up else output_geom.y_max
    dst_x_scale = output_geom.x_res
    dst_y_scale = output_geom.y_res \
        if output_geom.is_j_axis_up else -output_geom.y_res
    # TODO: GridMapping: this will cause out-of-memory problems!
    x_values, y_values = src_geo_coding.xy_coords.values
    _compute_ij_images_numpy_parallel(x_values,
                                      y_values,
                                      0,
                                      0,
                                      dst_src_ij_images,
                                      dst_x_offset,
                                      dst_y_offset,
                                      dst_x_scale,
                                      dst_y_scale,
                                      uv_delta)
    return dst_src_ij_images


def _compute_ij_images_xarray_dask(src_geo_coding: GridMapping,
                                   output_geom: GridMapping,
                                   uv_delta: float) -> da.Array:
    """
    Compute dask.array.Array destination image
    with source pixel i,j coords from xarray.DataArray x,y sources.
    """
    dst_width = output_geom.width
    dst_height = output_geom.height
    dst_tile_width = output_geom.tile_width
    dst_tile_height = output_geom.tile_height
    dst_var_shape = 2, dst_height, dst_width
    dst_var_chunks = 2, dst_tile_height, dst_tile_width

    dst_x_min, dst_y_min, dst_x_max, dst_y_max = output_geom.xy_bbox
    dst_x_res, dst_y_res = output_geom.xy_res
    dst_is_j_axis_up = output_geom.is_j_axis_up

    # Compute an empirical xy_border as a function of the
    # number of tiles, because the more tiles we have
    # the smaller the destination xy-bboxes and the higher
    # the risk to not find any source ij-bbox for a given xy-bbox.
    # xy_border will not be larger than half of the
    # coverage of a tile.
    #
    num_tiles_x = dst_width / dst_tile_width
    num_tiles_y = dst_height / dst_tile_height
    xy_border = min(min(2 * num_tiles_x * output_geom.x_res,
                        2 * num_tiles_y * output_geom.y_res),
                    min(0.5 * (dst_x_max - dst_x_min),
                        0.5 * (dst_y_max - dst_y_min)))

    dst_xy_bboxes = output_geom.xy_bboxes
    src_ij_bboxes = src_geo_coding.ij_bboxes_from_xy_bboxes(
        dst_xy_bboxes,
        xy_border=xy_border,
        ij_border=1
    )

    return compute_array_from_func(
        _compute_ij_images_xarray_dask_block,
        dst_var_shape,
        dst_var_chunks,
        np.float64,
        ctx_arg_names=[
            'dtype',
            'block_id',
            'block_shape',
            'block_slices',
        ],
        args=(
            src_geo_coding.xy_coords,
            src_ij_bboxes,
            dst_x_min,
            dst_y_min,
            dst_y_max,
            dst_x_res,
            dst_y_res,
            dst_is_j_axis_up,
            uv_delta
        ),
        name='ij_pixels'
    )


def _compute_ij_images_xarray_dask_block(
        dtype: np.dtype,
        block_id: int,
        block_shape: Tuple[int, int],
        block_slices: Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]],
        src_xy_coords: xr.DataArray,
        src_ij_bboxes: np.ndarray,
        dst_x_min: float,
        dst_y_min: float,
        dst_y_max: float,
        dst_x_res: float,
        dst_y_res: float,
        dst_is_j_axis_up: bool,
        uv_delta: float
) -> np.ndarray:
    """
    Compute dask.array.Array destination block with source
    pixel i,j coords from xarray.DataArray x,y sources.
    """
    dst_src_ij_block = np.full(block_shape, np.nan, dtype=dtype)
    _, (dst_y_slice_start, _), (dst_x_slice_start, _) = block_slices
    src_ij_bbox = src_ij_bboxes[block_id]
    src_i_min, src_j_min, src_i_max, src_j_max = src_ij_bbox
    if src_i_min == -1:
        return dst_src_ij_block
    src_xy_values = src_xy_coords[
                    :, src_j_min:src_j_max + 1, src_i_min:src_i_max + 1
                    ].values
    src_x_values = src_xy_values[0]
    src_y_values = src_xy_values[1]
    dst_x_offset = dst_x_min + dst_x_slice_start * dst_x_res
    if dst_is_j_axis_up:
        dst_y_offset = dst_y_min + dst_y_slice_start * dst_y_res
    else:
        dst_y_offset = dst_y_max - dst_y_slice_start * dst_y_res
    _compute_ij_images_numpy_sequential(
        src_x_values,
        src_y_values,
        src_i_min,
        src_j_min,
        dst_src_ij_block,
        dst_x_offset,
        dst_y_offset,
        dst_x_res,
        dst_y_res if dst_is_j_axis_up else -dst_y_res,
        uv_delta
    )
    return dst_src_ij_block


@nb.njit(nogil=True, parallel=True, cache=True)
def _compute_ij_images_numpy_parallel(
        src_x_image: np.ndarray,
        src_y_image: np.ndarray,
        src_i_min: int,
        src_j_min: int,
        dst_src_ij_images: np.ndarray,
        dst_x_offset: float,
        dst_y_offset: float,
        dst_x_scale: float,
        dst_y_scale: float,
        uv_delta: float
):
    """
    Compute numpy.ndarray destination image with source
    pixel i,j coords from numpy.ndarray x,y sources in
    parallel mode.
    """
    src_height = src_x_image.shape[-2]
    dst_src_ij_images[:, :, :] = np.nan
    for src_j0 in nb.prange(src_height - 1):
        _compute_ij_images_for_source_line(
            src_j0,
            src_x_image,
            src_y_image,
            src_i_min,
            src_j_min,
            dst_src_ij_images,
            dst_x_offset,
            dst_y_offset,
            dst_x_scale,
            dst_y_scale,
            uv_delta
        )


# Extra dask version, because if we use parallel=True
# and nb.prange, we end up in infinite JIT compilation :(
@nb.njit(nogil=True, cache=True)
def _compute_ij_images_numpy_sequential(
        src_x_image: np.ndarray,
        src_y_image: np.ndarray,
        src_i_min: int,
        src_j_min: int,
        dst_src_ij_images: np.ndarray,
        dst_x_offset: float,
        dst_y_offset: float,
        dst_x_scale: float,
        dst_y_scale: float,
        uv_delta: float
):
    """
    Compute numpy.ndarray destination image with source pixel i,j coords
    from numpy.ndarray x,y sources NOT in parallel mode.
    """
    src_height = src_x_image.shape[-2]
    dst_src_ij_images[:, :, :] = np.nan
    for src_j0 in range(src_height - 1):
        _compute_ij_images_for_source_line(
            src_j0,
            src_x_image,
            src_y_image,
            src_i_min,
            src_j_min,
            dst_src_ij_images,
            dst_x_offset,
            dst_y_offset,
            dst_x_scale,
            dst_y_scale,
            uv_delta
        )


@nb.njit(nogil=True, cache=True)
def _compute_ij_images_for_source_line(
        src_j0: int,
        src_x_image: np.ndarray,
        src_y_image: np.ndarray,
        src_i_min: int,
        src_j_min: int,
        dst_src_ij_images: np.ndarray,
        dst_x_offset: float,
        dst_y_offset: float,
        dst_x_scale: float,
        dst_y_scale: float,
        uv_delta: float
):
    """
    Compute numpy.ndarray destination image with source
    pixel i,j coords from a numpy.ndarray x,y source line.
    """
    src_width = src_x_image.shape[-1]

    dst_width = dst_src_ij_images.shape[-1]
    dst_height = dst_src_ij_images.shape[-2]

    dst_px = np.zeros(4, dtype=src_x_image.dtype)
    dst_py = np.zeros(4, dtype=src_y_image.dtype)

    u_min = v_min = -uv_delta
    uv_max = 1.0 + 2 * uv_delta

    for src_i0 in range(src_width - 1):
        src_i1 = src_i0 + 1
        src_j1 = src_j0 + 1

        dst_px[0] = dst_p0x = src_x_image[src_j0, src_i0]
        dst_px[1] = dst_p1x = src_x_image[src_j0, src_i1]
        dst_px[2] = dst_p2x = src_x_image[src_j1, src_i0]
        dst_px[3] = dst_p3x = src_x_image[src_j1, src_i1]

        dst_py[0] = dst_p0y = src_y_image[src_j0, src_i0]
        dst_py[1] = dst_p1y = src_y_image[src_j0, src_i1]
        dst_py[2] = dst_p2y = src_y_image[src_j1, src_i0]
        dst_py[3] = dst_p3y = src_y_image[src_j1, src_i1]

        dst_pi = np.floor((dst_px - dst_x_offset)
                          / dst_x_scale).astype(np.int64)
        dst_pj = np.floor((dst_py - dst_y_offset)
                          / dst_y_scale).astype(np.int64)

        dst_i_min = np.min(dst_pi)
        dst_i_max = np.max(dst_pi)
        dst_j_min = np.min(dst_pj)
        dst_j_max = np.max(dst_pj)

        if dst_i_max < 0 \
                or dst_j_max < 0 \
                or dst_i_min >= dst_width \
                or dst_j_min >= dst_height:
            continue

        if dst_i_min < 0:
            dst_i_min = 0

        if dst_i_max >= dst_width:
            dst_i_max = dst_width - 1

        if dst_j_min < 0:
            dst_j_min = 0

        if dst_j_max >= dst_height:
            dst_j_max = dst_height - 1

        # u from p0 right to p1, v from p0 down to p2
        det_a = _fdet(dst_p0x, dst_p0y, dst_p1x, dst_p1y, dst_p2x, dst_p2y)
        # u from p3 left to p2, v from p3 up to p1
        det_b = _fdet(dst_p3x, dst_p3y, dst_p2x, dst_p2y, dst_p1x, dst_p1y)

        if np.isnan(det_a) or np.isnan(det_b):
            # print('no plane at:', src_i0, src_j0)
            continue

        for dst_j in range(dst_j_min, dst_j_max + 1):
            dst_y = dst_y_offset + (dst_j + 0.5) * dst_y_scale
            for dst_i in range(dst_i_min, dst_i_max + 1):
                dst_x = dst_x_offset + (dst_i + 0.5) * dst_x_scale

                # TODO: use two other combinations,
                #       if one of the dst_px<n>,dst_py<n> pairs is missing.

                src_i = src_j = -1

                if det_a != 0.0:
                    u = _fu(dst_x, dst_y,
                            dst_p0x, dst_p0y, dst_p2x, dst_p2y) / det_a
                    v = _fv(dst_x, dst_y,
                            dst_p0x, dst_p0y, dst_p1x, dst_p1y) / det_a
                    if u >= u_min and v >= v_min and u + v <= uv_max:
                        src_i = src_i0 + _fclamp(u, 0.0, 1.0)
                        src_j = src_j0 + _fclamp(v, 0.0, 1.0)
                if src_i == -1 and det_b != 0.0:
                    u = _fu(dst_x, dst_y,
                            dst_p3x, dst_p3y, dst_p1x, dst_p1y) / det_b
                    v = _fv(dst_x, dst_y,
                            dst_p3x, dst_p3y, dst_p2x, dst_p2y) / det_b
                    if u >= u_min and v >= v_min and u + v <= uv_max:
                        src_i = src_i1 - _fclamp(u, 0.0, 1.0)
                        src_j = src_j1 - _fclamp(v, 0.0, 1.0)
                if src_i != -1:
                    dst_src_ij_images[0, dst_j, dst_i] = src_i_min + src_i
                    dst_src_ij_images[1, dst_j, dst_i] = src_j_min + src_j


def _compute_var_image_xarray_numpy(
        src_var: xr.DataArray,
        dst_src_ij_images: np.ndarray,
        fill_value: Union[int, float, complex] = np.nan
) -> np.ndarray:
    """
    Extract source pixels from xarray.DataArray source
    with numpy.ndarray data.
    """
    return _compute_var_image_numpy(src_var.values,
                                    dst_src_ij_images,
                                    fill_value)


def _compute_var_image_xarray_dask(
        src_var: xr.DataArray,
        dst_src_ij_images: np.ndarray,
        fill_value: Union[int, float, complex] = np.nan
) -> da.Array:
    """
    Extract source pixels from xarray.DataArray source
    with dask.array.Array data.
    """
    return da.map_blocks(_compute_var_image_xarray_dask_block,
                         src_var.values,
                         dst_src_ij_images,
                         fill_value,
                         dtype=src_var.dtype,
                         drop_axis=0)


@nb.njit(nogil=True, cache=True)
def _compute_var_image_numpy(
        src_var: np.ndarray,
        dst_src_ij_images: np.ndarray,
        fill_value: Union[int, float, complex] = np.nan
) -> np.ndarray:
    """
    Extract source pixels from numpy.ndarray source
    with numba in parallel mode.
    """
    dst_width = dst_src_ij_images.shape[-1]
    dst_height = dst_src_ij_images.shape[-2]
    dst_shape = src_var.shape[:-2] + (dst_height, dst_width)
    dst_values = np.full(dst_shape, fill_value, dtype=src_var.dtype)
    _compute_var_image_numpy_parallel(src_var,
                                      dst_src_ij_images,
                                      dst_values)
    return dst_values


@nb.njit(nogil=True, cache=True)
def _compute_var_image_xarray_dask_block(
        src_var_image: np.ndarray,
        dst_src_ij_images: np.ndarray,
        fill_value: Union[int, float, complex] = np.nan
) -> np.ndarray:
    """
    Extract source pixels from np.ndarray source
    and return a block of a dask array.
    """
    dst_width = dst_src_ij_images.shape[-1]
    dst_height = dst_src_ij_images.shape[-2]
    dst_shape = src_var_image.shape[:-2] + (dst_height, dst_width)
    dst_values = np.full(dst_shape, fill_value, dtype=src_var_image.dtype)
    _compute_var_image_numpy_sequential(src_var_image,
                                        dst_src_ij_images,
                                        dst_values)
    return dst_values


@nb.njit(nogil=True, parallel=True, cache=True)
def _compute_var_image_numpy_parallel(
        src_var_image: np.ndarray,
        dst_src_ij_images: np.ndarray,
        dst_var_image: np.ndarray
):
    """
    Extract source pixels from np.ndarray source
    using numba parallel mode.
    """
    dst_height = dst_var_image.shape[-2]
    for dst_j in nb.prange(dst_height):
        _compute_var_image_for_dest_line(
            dst_j, src_var_image, dst_src_ij_images, dst_var_image
        )


# Extra dask version, because if we use parallel=True
# and nb.prange, we end up in infinite JIT compilation :(
@nb.njit(nogil=True, cache=True)
def _compute_var_image_numpy_sequential(
        src_var_image: np.ndarray,
        dst_src_ij_images: np.ndarray,
        dst_var_image: np.ndarray
):
    """
    Extract source pixels from np.ndarray source
    NOT using numba parallel mode.
    """
    dst_height = dst_var_image.shape[-2]
    for dst_j in range(dst_height):
        _compute_var_image_for_dest_line(
            dst_j, src_var_image, dst_src_ij_images, dst_var_image
        )


@nb.njit(nogil=True, cache=True)
def _compute_var_image_for_dest_line(
        dst_j: int,
        src_var_image: np.ndarray,
        dst_src_ij_images: np.ndarray,
        dst_var_image: np.ndarray
):
    """
    Extract source pixels from *src_values* np.ndarray
    and write into dst_values np.ndarray.
    """
    src_width = src_var_image.shape[-1]
    src_height = src_var_image.shape[-2]
    dst_width = dst_var_image.shape[-1]
    src_i_min = 0
    src_j_min = 0
    src_i_max = src_width - 1
    src_j_max = src_height - 1
    for dst_i in range(dst_width):
        src_i_f = dst_src_ij_images[0, dst_j, dst_i]
        src_j_f = dst_src_ij_images[1, dst_j, dst_i]
        if np.isnan(src_i_f) or np.isnan(src_j_f):
            continue
        # Note int() is 2x faster than math.floor() and should yield the same results for only positive i,j.
        src_i = int(src_i_f)
        src_j = int(src_j_f)
        u = src_i_f - src_i
        v = src_j_f - src_j
        if u > 0.5:
            src_i = _iclamp(src_i + 1, src_i_min, src_i_max)
        if v > 0.5:
            src_j = _iclamp(src_j + 1, src_j_min, src_j_max)
        dst_var_image[..., dst_j, dst_i] = src_var_image[..., src_j, src_i]


@nb.njit('float64(float64, float64, float64, float64, float64, float64)',
         nogil=True, inline='always')
def _fdet(px0: float, py0: float,
          px1: float, py1: float,
          px2: float, py2: float) -> float:
    return (px0 - px1) * (py0 - py2) - (px0 - px2) * (py0 - py1)


@nb.njit('float64(float64, float64, float64, float64, float64, float64)',
         nogil=True, inline='always')
def _fu(px: float, py: float,
        px0: float, py0: float,
        px2: float, py2: float) -> float:
    return (px0 - px) * (py0 - py2) - (py0 - py) * (px0 - px2)


@nb.njit('float64(float64, float64, float64, float64, float64, float64)',
         nogil=True, inline='always')
def _fv(px: float, py: float,
        px0: float, py0: float,
        px1: float, py1: float) -> float:
    return (py0 - py) * (px0 - px1) - (px0 - px) * (py0 - py1)


@nb.njit('float64(float64, float64, float64)', nogil=True, inline='always')
def _fclamp(x: float, x_min: float, x_max: float) -> float:
    return x_min if x < x_min else (x_max if x > x_max else x)


@nb.njit('int64(int64, int64, int64)', nogil=True, inline='always')
def _iclamp(x: int, x_min: int, x_max: int) -> int:
    return x_min if x < x_min else (x_max if x > x_max else x)


def _millis(seconds: float) -> int:
    return round(1000 * seconds)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import abstractmethod, ABC

import xarray as xr

from xcube.util.jsonschema import JsonObjectSchema

# Name of the (instance) method that processes a
# dataset and returns a new one.
# The one and only argument of the method is an xarray.Dataset.
# Then any number of processing parameters can be passed as
# keyword-arguments.
# See :meth:DatasetProcessor#process_dataset.
METHOD_NAME_DATASET_PROCESSOR = 'process_dataset'

# Name of the (static/class) method that returns
# the JSON Object Schema for the processor parameters.
# The methods takes no arguments.
# See :meth:DatasetProcessor#get_process_params_schema.
METHOD_NAME_PARAMS_SCHEMA_GETTER = 'get_process_params_schema'


class DatasetProcessor(ABC):
    """
    A generic dataset processor.
    """

    @abstractmethod
    def process_dataset(self,
                        dataset: xr.Dataset,
                        **process_params) -> xr.Dataset:
        """
        Process *dataset* into a new dataset using the
        given processing parameters *process_params*.

        The passed *process_params* must validate against
        the JSON schema returned by the :meth:get_process_params_schema
        class method.

        :param dataset: The dataset to be processed.
        :param process_params: The processing parameters.
        :return: A new dataset.
        """

    @classmethod
    def get_process_params_schema(cls) -> JsonObjectSchema:
        """
        Get the JSON Schema for the processing parameters passed
        to :meth:process_dataset.

        The parameters are described by the JSON Object Schema's
        properties.

        The default implementation returns a JSON Schema
        that allows for any parameters of any type.

        :return: a JSON Object Schema.
        """
        return JsonObjectSchema(additional_properties=True)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import Dict, Any, Sequence, Union

import numpy as np
import xarray as xr

from xcube.core.schema import CubeSchema
from xcube.core.select import select_variables_subset
from xcube.core.verify import assert_cube


def resample_in_time(dataset: xr.Dataset,
                     frequency: str,
                     method: Union[str, Sequence[str]],
                     offset=None,
                     base=None,
                     tolerance=None,
                     interp_kind=None,
                     time_chunk_size=None,
                     var_names: Sequence[str] = None,
                     metadata: Dict[str, Any] = None,
                     cube_asserted: bool = False) -> xr.Dataset:
    """
    Resample a dataset in the time dimension.

    The argument *method* may be one or a sequence of
    ``'all'``, ``'any'``,
    ``'argmax'``, ``'argmin'``, ``'count'``,
    ``'first'``, ``'last'``,
    ``'max'``, ``'min'``, ``'mean'``, ``'median'``,
    ``'percentile_<p>'``,
    ``'std'``, ``'sum'``, ``'var'``.

    In value ``'percentile_<p>'`` is a placeholder,
    where ``'<p>'`` must be replaced by an integer percentage
    value, e.g. ``'percentile_90'`` is the 90%-percentile.

    *Important note:* As of xarray 0.14 and dask 2.8, the
    methods ``'median'`` and ``'percentile_<p>'` cannot be
    used if the variables in *cube* comprise chunked dask arrays.
    In this case, use the ``compute()`` or ``load()`` method
    to convert dask arrays into numpy arrays.

    :param dataset: The xcube dataset.
    :param frequency: Temporal aggregation frequency.
        Use format "<count><offset>" where <offset> is one of
        'H', 'D', 'W', 'M', 'Q', 'Y'.
    :param method: Resampling method or sequence of
        resampling methods.
    :param offset: Offset used to adjust the resampled time labels.
        Uses same syntax as *frequency*.
    :param base: Deprecated since xcube 1.0.4.
        No longer used as of pandas 2.0.
    :param time_chunk_size: If not None, the chunk size to be
        used for the "time" dimension.
    :param var_names: Variable names to include.
    :param tolerance: Time tolerance for selective
        upsampling methods. Defaults to *frequency*.
    :param interp_kind: Kind of interpolation
        if *method* is 'interpolation'.
    :param metadata: Output metadata.
    :param cube_asserted: If False, *cube* will be verified,
        otherwise it is expected to be a valid cube.
    :return: A new xcube dataset resampled in time.
    """
    if not cube_asserted:
        assert_cube(dataset)

    if base is not None:
        warnings.warn("Keyword 'base' is deprecated and no longer used.")

    if frequency == 'all':
        time_gap = np.array(dataset.time[-1]) - np.array(dataset.time[0])
        days = int((np.timedelta64(time_gap, 'D')
                    / np.timedelta64(1, 'D')) + 1)
        frequency = f'{days}D'

    if var_names:
        dataset = select_variables_subset(dataset, var_names)

    resampler = dataset.resample(skipna=True,
                                 closed='left',
                                 label='left',
                                 time=frequency,
                                 loffset=offset)

    if isinstance(method, str):
        methods = [method]
    else:
        methods = list(method)

    percentile_prefix = 'percentile_'

    resampled_cubes = []
    for method in methods:
        method_args = []
        method_postfix = method
        if method.startswith(percentile_prefix):
            p = int(method[len(percentile_prefix):])
            q = p / 100.0
            method_args = [q]
            method_postfix = f'p{p}'
            method = 'quantile'
        resampling_method = getattr(resampler, method)
        method_kwargs = get_method_kwargs(method,
                                          frequency,
                                          interp_kind,
                                          tolerance)
        resampled_cube = resampling_method(*method_args,
                                           **method_kwargs)
        resampled_cube = resampled_cube.rename(
            {var_name: f'{var_name}_{method_postfix}'
             for var_name in resampled_cube.data_vars})
        resampled_cubes.append(resampled_cube)

    if len(resampled_cubes) == 1:
        resampled_cube = resampled_cubes[0]
    else:
        resampled_cube = xr.merge(resampled_cubes)

    # TODO: add time_bnds to resampled_ds
    time_coverage_start = '%s' % dataset.time[0]
    time_coverage_end = '%s' % dataset.time[-1]

    resampled_cube.attrs.update(metadata or {})
    # TODO: add other time_coverage_ attributes
    resampled_cube.attrs.update(time_coverage_start=time_coverage_start,
                                time_coverage_end=time_coverage_end)

    schema = CubeSchema.new(dataset)
    chunk_sizes = {schema.dims[i]: schema.chunks[i] for i in range(schema.ndim)}

    if isinstance(time_chunk_size, int) and time_chunk_size >= 0:
        chunk_sizes['time'] = time_chunk_size

    return resampled_cube.chunk(chunk_sizes)


def get_method_kwargs(method, frequency, interp_kind, tolerance):
    if method == 'interpolate':
        kwargs = {'kind': interp_kind or 'linear'}
    elif method in {'nearest', 'bfill', 'ffill', 'pad'}:
        kwargs = {'tolerance': tolerance or frequency}
    elif method in {'first', 'last', 'sum',
                    'min', 'max',
                    'mean', 'median', 'std', 'var'}:
        kwargs = {'dim': 'time', 'keep_attrs': True, 'skipna': True}
    elif method == 'prod':
        kwargs = {'dim': 'time', 'skipna': True}
    elif method == 'count':
        kwargs = {'dim': 'time', 'keep_attrs': True}
    else:
        kwargs = {}
    return kwargs



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
from abc import abstractmethod, ABC
from typing import Generic, TypeVar, Tuple, Any, Iterable, Dict, List, \
    Sequence, Mapping

K = TypeVar('K')
V = TypeVar('V')

_DICT_IS_READONLY = 'dict is read-only'
_LIST_IS_READONLY = 'list is read-only'


class Frozen(ABC):
    """A frozen, that is, an immutable object."""

    @classmethod
    @abstractmethod
    def freeze(cls, obj: Any) -> "Frozen":
        """Deeply freeze object *obj*, i.e.,
        return an immutable version of it.
        """

    @abstractmethod
    def defrost(self) -> Any:
        """Deeply defrost this frozen object, i.e.,
        return its mutable counterpart.
        """


class FrozenDict(Dict[K, V], Frozen, Generic[K, V]):
    """A frozen version of a standard ``dict``."""

    @classmethod
    def freeze(cls, other: Mapping) -> "FrozenDict":
        return FrozenDict({key: freeze_value(value)
                           for key, value in other.items()})

    def defrost(self) -> Dict[K, V]:
        return {key: defrost_value(value)
                for key, value in self.items()}

    ###########################################################
    # dict overrides

    def clear(self) -> None:
        raise TypeError(_DICT_IS_READONLY)

    def pop(self, *args, **kwargs) -> V:
        raise TypeError(_DICT_IS_READONLY)

    def popitem(self) -> Tuple[K, V]:
        raise TypeError(_DICT_IS_READONLY)

    def update(self, *args, **kwargs) -> None:
        raise TypeError(_DICT_IS_READONLY)

    def __setitem__(self, key: K, value: V) -> None:
        raise TypeError(_DICT_IS_READONLY)

    def __delitem__(self, key: K) -> None:
        raise TypeError(_DICT_IS_READONLY)

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError(_DICT_IS_READONLY)

    def __delattr__(self, name: str) -> None:
        raise TypeError(_DICT_IS_READONLY)


class FrozenList(List[V], Frozen, Generic[V]):
    """A frozen version of a standard ``list``."""

    @classmethod
    def freeze(cls, other: Sequence[V]) -> "FrozenList":
        return FrozenList(freeze_value(item) for item in other)

    def defrost(self) -> List[V]:
        return [defrost_value(item) for item in self]

    ###########################################################
    # list overrides

    def append(self, element: V):
        raise TypeError(_LIST_IS_READONLY)

    def extend(self, elements: Iterable[V]):
        raise TypeError(_LIST_IS_READONLY)

    def clear(self):
        raise TypeError(_LIST_IS_READONLY)

    def insert(self, index: int, elements: Iterable[V]):
        raise TypeError(_LIST_IS_READONLY)

    def pop(self, *args, **kwargs):
        raise TypeError(_LIST_IS_READONLY)

    def remove(self, value: V):
        raise TypeError(_LIST_IS_READONLY)

    def reverse(self):
        raise TypeError(_LIST_IS_READONLY)

    def sort(self, *args, **kwargs):
        raise TypeError(_LIST_IS_READONLY)

    def __iadd__(self, arg: Any):
        raise TypeError(_LIST_IS_READONLY)

    def __imul__(self, arg: Any):
        raise TypeError(_LIST_IS_READONLY)

    def __setitem__(self, index: int, value: V):
        raise TypeError(_LIST_IS_READONLY)


# Standard Python immutables
_IMMUTABLES = type(None), bool, int, float, complex, str, bytes


def freeze_value(value: Any) -> Any:
    """Freeze given *value*, that is, return a deeply
    immutable version of it."""
    # Note, _IMMUTABLES also includes str and bytes which are
    # both a collections.abc.Sequence.
    if isinstance(value, _IMMUTABLES):
        return value
    if isinstance(value, collections.abc.Mapping):
        return FrozenDict.freeze(value)
    if isinstance(value, collections.abc.Sequence):
        return FrozenList.freeze(value)
    return value


def defrost_value(value: Any) -> Any:
    """Defrost given *value*, that is, return a deeply
    mutable version of it."""
    if isinstance(value, Frozen):
        return value.defrost()
    return value



from typing import Dict, TypeVar, Optional


def caseless_dict(*args, **kwargs) -> Dict:
    """Create a dictionary that compares its string keys in a case-insensitive manner."""
    return _CaselessDict(*args, **kwargs)


_KT = TypeVar('_KT')
_VT = TypeVar('_VT')
_VT_co = TypeVar('_VT_co', covariant=True)


class _CaselessDict(dict):

    def __init__(self, *args, **kwargs: _VT):
        super().__init__(*args, **kwargs)
        self._lc_keys = {k.lower() if isinstance(k, str) else k: k for k in self.keys()}

    def __getitem__(self, k: _KT) -> _VT:
        if isinstance(k, str):
            k = self._lc_keys.get(k.lower(), k)
        return super().__getitem__(k)

    def __setitem__(self, k: _KT, v: _VT) -> None:
        if isinstance(k, str):
            lc_key = k.lower()
            if lc_key in self._lc_keys:
                k = self._lc_keys[lc_key]
            else:
                self._lc_keys[lc_key] = k
        super().__setitem__(k, v)

    def __delitem__(self, k: _KT) -> None:
        if isinstance(k, str):
            lc_key = k.lower()
            if lc_key in self._lc_keys:
                k = self._lc_keys.pop(lc_key)
        super().__delitem__(k)

    def __contains__(self, k: object) -> bool:
        if isinstance(k, str):
            return k.lower() in self._lc_keys
        return super().__contains__(k)

    def get(self, k: _KT, default: _VT = None) -> Optional[_VT_co]:
        if isinstance(k, str):
            lc_key = k.lower()
            if lc_key in self._lc_keys:
                k = self._lc_keys[lc_key]
            else:
                return default
        return super().get(k, default) if default is not None else super().get(k)

    def pop(self, k: _KT) -> _VT:
        if isinstance(k, str):
            lc_key = k.lower()
            if lc_key in self._lc_keys:
                k = self._lc_keys.pop(lc_key)
        return super().pop(k)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import json
from typing import Any, Union, List, Dict

import numpy as np

JsonArray = List["JsonValue"]
JsonObject = Dict[str, "JsonValue"]
JsonValue = Union[None, bool, int, float, str, JsonArray, JsonObject]


class NumpyJSONEncoder(json.JSONEncoder):
    """A JSON encoder that converts numpy-like
    scalars into corresponding serializable Python objects.
    """

    def default(self, obj: Any) -> JsonValue:
        converted_obj = _convert_default(obj)
        if converted_obj is not obj:
            return converted_obj
        return json.JSONEncoder.default(self, obj)


_PRIMITIVE_JSON_TYPES = {
    type(None),
    bool,
    int,
    float,
    str
}


def to_json_value(obj: Any) -> JsonValue:
    """Convert *obj* into a JSON-serializable object.

    :param obj: A Python object.
    :return: A JSON-serializable version of *obj*, or *obj*
        if *obj* is already JSON-serializable.
    :raises TypeError: If *obj* cannot be made JSON-serializable
    """
    if obj is None:
        return None

    converted_obj = _convert_default(obj)
    if converted_obj is not obj:
        return converted_obj

    obj_type = type(obj)

    if obj_type in _PRIMITIVE_JSON_TYPES:
        return obj

    for t in _PRIMITIVE_JSON_TYPES:
        if isinstance(obj, t):
            return t(obj)

    if obj_type is dict:
        converted_obj = {_key(k): to_json_value(v) for k, v in obj.items()}
        if any(converted_obj[k] is not obj[k] for k in obj.keys()):
            return converted_obj
        else:
            return obj

    if obj_type is list:
        converted_obj = [to_json_value(item) for item in obj]
        if any(o1 is not o2 for o1, o2 in zip(converted_obj, obj)):
            return converted_obj
        else:
            return obj

    try:
        return {_key(k): to_json_value(v) for k, v in obj.items()}
    except AttributeError:
        try:
            return [to_json_value(item) for item in obj]
        except TypeError:
            # Same as json.JSONEncoder.default(self, obj)
            raise TypeError(f'Object of type'
                            f' {obj.__class__.__name__}'
                            f' is not JSON serializable')


def _key(key: Any) -> str:
    if not isinstance(key, str):
        raise TypeError(f'Property names of JSON objects must be strings,'
                        f' but got {key.__class__.__name__}')
    return key


def _convert_default(obj: Any) -> Any:
    if hasattr(obj, 'dtype') and hasattr(obj, 'ndim'):
        if obj.ndim == 0:
            if np.issubdtype(obj.dtype, bool):
                return bool(obj)
            elif np.issubdtype(obj.dtype, np.integer):
                return int(obj)
            elif np.issubdtype(obj.dtype, np.floating):
                return float(obj)
            elif np.issubdtype(obj.dtype, np.datetime64):
                return np.datetime_as_string(obj, timezone='UTC', unit='s')
            elif np.issubdtype(obj.dtype, np.str):
                return str(obj)
        else:
            return [_convert_default(item) for item in obj]
    # We may handle other non-JSON-serializable datatypes here
    return obj



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import threading
import time
import traceback
from abc import ABC
from typing import Sequence, Optional, Any, Tuple, Type, List

import dask.callbacks
import dask.diagnostics

from xcube.util.assertions import assert_true, assert_given


class ProgressState:
    """Represents the state of progress."""

    def __init__(self, label: str, total_work: float, super_work: float):
        self._label = label
        self._total_work = total_work
        self._super_work = super_work
        self._super_work_ahead = 1.
        self._exc_info = None
        self._traceback = None
        self._completed_work = 0.
        self._finished = False
        self._start_time = None
        self._start_time = time.perf_counter()
        self._total_time = None

    @property
    def label(self) -> str:
        return self._label

    @property
    def total_work(self) -> float:
        return self._total_work

    @property
    def super_work(self) -> float:
        return self._super_work

    @property
    def completed_work(self) -> float:
        return self._completed_work

    @property
    def progress(self) -> float:
        return self._completed_work / self._total_work

    def to_super_work(self, work: float) -> float:
        return self._super_work * work / self._total_work

    @property
    def exc_info(self) -> Optional[Tuple[Type, BaseException, Any]]:
        return self._exc_info

    @exc_info.setter
    def exc_info(self, exc_info: Tuple[Type, BaseException, Any]):
        self._exc_info = exc_info

    @property
    def exc_info_text(self) -> Optional[Tuple[str, str, List[str]]]:
        if not self.exc_info:
            return None
        exc_type, exc_value, exc_traceback = self.exc_info
        return (f'{type(exc_value).__name__}',
                f'{exc_value}',
                traceback.format_exception(exc_type, exc_value, exc_traceback))

    @property
    def finished(self) -> bool:
        return self._finished

    @property
    def total_time(self) -> Optional[float]:
        return self._total_time

    @property
    def super_work_ahead(self) -> float:
        return self._super_work_ahead

    @super_work_ahead.setter
    def super_work_ahead(self, work: float):
        assert_true(work > 0, 'work must be greater than zero')
        self._super_work_ahead = work

    def inc_work(self, work: float):
        assert_true(work > 0, 'work must be greater than zero')
        self._completed_work += work

    def finish(self):
        self._finished = True
        self._total_time = time.perf_counter() - self._start_time


class ProgressObserver(ABC):
    """
    A progress observer is notified about nested state changes when using the
    :class:observe_progress context manager.
    """

    def on_begin(self, state_stack: Sequence[ProgressState]):
        """
        Called, if an observed code block begins execution.
        """

    def on_update(self, state_stack: Sequence[ProgressState]):
        """
        Called, if the progress state has changed within an observed code block.
        """

    def on_end(self, state_stack: Sequence[ProgressState]):
        """
        Called, if an observed block of code ends execution.
        """

    def activate(self):
        _ProgressContext.instance().add_observer(self)

    def deactivate(self):
        _ProgressContext.instance().remove_observer(self)


class _ProgressContext:
    _instance = None

    def __init__(self, *observers: ProgressObserver):
        self._observers = set(observers)
        self._state_stack = list()

    def add_observer(self, observer: ProgressObserver):
        self._observers.add(observer)

    def remove_observer(self, observer: ProgressObserver):
        self._observers.discard(observer)

    def emit_begin(self):
        for observer in self._observers:
            observer.on_begin(self._state_stack)

    def emit_update(self):
        for observer in self._observers:
            observer.on_update(self._state_stack)

    def emit_end(self):
        for observer in self._observers:
            observer.on_end(self._state_stack)

    def begin(self, label: str, total_work: float) -> ProgressState:
        super_work = self._state_stack[-1].super_work_ahead \
            if self._state_stack else 1
        progress_state = ProgressState(label, total_work, super_work)
        self._state_stack.append(progress_state)
        self.emit_begin()
        return progress_state

    def end(self, exc_type, exc_value, exc_traceback) -> ProgressState:
        exc_info = tuple((exc_type, exc_value, exc_traceback))
        progress_state = self._state_stack[-1]
        progress_state.exc_info = exc_info if any(exc_info) else None
        progress_state.finish()
        self.emit_end()
        self._state_stack.pop()
        if self._state_stack:
            self._state_stack[-1].super_work_ahead = 1
        return progress_state

    def worked(self, work: float):
        assert_true(self._state_stack,
                    'worked() method call is missing a current context')
        assert_true(work > 0, 'work must be greater than zero')
        for s in reversed(self._state_stack):
            s.inc_work(work)
            work = s.to_super_work(work)
        self.emit_update()

    def will_work(self, work: float):
        assert_true(self._state_stack,
                    'will_work() method call is missing a current context')
        # noinspection PyProtectedMember
        self._state_stack[-1].super_work_ahead = work

    @classmethod
    def instance(cls) -> '_ProgressContext':
        return cls._instance

    @classmethod
    def set_instance(cls, instance: '_ProgressContext' = None) \
            -> '_ProgressContext':
        cls._instance, old_instance = \
            (instance or _ProgressContext()), cls._instance
        return old_instance


_ProgressContext.set_instance()


class new_progress_observers:
    """
    Takes zero or more progress observers and activates them in the enclosed context.
    Progress observers from an outer context will no longer be active.

    :param observers: progress observers that will temporarily replace existing ones.
    """

    def __init__(self, *observers: ProgressObserver):
        self._observers = observers
        self._old_context = None

    def __enter__(self):
        self._old_context = _ProgressContext.set_instance(
            _ProgressContext(*self._observers)
        )

    def __exit__(self, type, value, traceback):
        _ProgressContext.set_instance(self._old_context)


class add_progress_observers:
    """
    Takes zero or more progress observers and uses them only in the
    enclosed context. Any progress observers from an outer context
    remain active.

    :param observers: progress observers to be added temporarily.
    """

    def __init__(self, *observers: ProgressObserver):
        self._observers = observers

    def __enter__(self):
        for observer in self._observers:
            observer.activate()

    def __exit__(self, type, value, traceback):
        for observer in self._observers:
            observer.deactivate()


class observe_progress:
    """
    Context manager for observing progress in the enclosed context.

    :param label: A label.
    :param total_work: The total work.
    """

    def __init__(self, label: str, total_work: float):
        assert_given(label, 'label')
        assert_true(total_work > 0, 'total_work must be greater than zero')
        self._label = label
        self._total_work = total_work
        self._state: Optional[ProgressState] = None

    @property
    def label(self) -> str:
        return self._label

    @property
    def total_work(self) -> float:
        return self._total_work

    @property
    def state(self) -> ProgressState:
        self._assert_used_correctly()
        return self._state

    def __enter__(self) -> 'observe_progress':
        self._state = _ProgressContext.instance().begin(self._label,
                                                        self._total_work)
        return self

    def __exit__(self, type, value, traceback):
        _ProgressContext.instance().end(type, value, traceback)

    # noinspection PyMethodMayBeStatic
    def worked(self, work: float):
        self._assert_used_correctly()
        _ProgressContext.instance().worked(work)

    # noinspection PyMethodMayBeStatic
    def will_work(self, work: float):
        self._assert_used_correctly()
        _ProgressContext.instance().will_work(work)

    def _assert_used_correctly(self):
        assert_true(self._state is not None,
                    'observe_progress() must be used with "with" statement')


class observe_dask_progress(dask.callbacks.Callback):
    """
    Observe progress made by Dask tasks.

    :param label: A label.
    :param total_work: The total work.
    :param interval: Time in seconds to between progress reports.
    :param initial_interval: Time in seconds to wait before progress
        is reported.
    """

    def __init__(self,
                 label: str,
                 total_work: float,
                 interval: float = 0.1,
                 initial_interval: float = 0):
        super().__init__()
        assert_given(label, 'label')
        assert_true(total_work > 0, 'total_work must be greater than zero')
        self._label = label
        self._total_work = total_work
        self._state: Optional[ProgressState] = None
        self._initial_interval = initial_interval
        self._interval = interval
        self._last_worked = 0
        self._running = False

    def __enter__(self) -> 'observe_dask_progress':
        super().__enter__()
        self._state = _ProgressContext.instance().begin(self._label,
                                                        self._total_work)
        return self

    def __exit__(self, type, value, traceback):
        self._stop_thread()
        _ProgressContext.instance().end(type, value, traceback)
        super().__exit__(type, value, traceback)

    # noinspection PyUnusedLocal
    def _start(self, dsk):
        """Dask callback implementation."""
        self._dask_state = None
        self._start_time = time.perf_counter()
        # Start background thread
        self._running = True
        self._timer = threading.Thread(target=self._timer_func)
        self._timer.daemon = True
        self._timer.start()

    # noinspection PyUnusedLocal
    def _pretask(self, key, dsk, state):
        """Dask callback implementation."""
        self._dask_state = state

    # noinspection PyUnusedLocal
    def _posttask(self, key, result, dsk, state, worker_id):
        """Dask callback implementation."""
        self._update()

    # noinspection PyUnusedLocal
    def _finish(self, dsk, state, errored):
        """Dask callback implementation."""
        self._stop_thread()
        elapsed = time.perf_counter() - self._start_time
        if elapsed > self._initial_interval:
            self._update()

    def _timer_func(self):
        """Background thread for updating"""
        while self._running:
            elapsed = time.perf_counter() - self._start_time
            if elapsed > self._initial_interval:
                self._update()
            time.sleep(self._interval)

    def _update(self):
        dask_state = self._dask_state
        if not dask_state:
            return
        num_done = len(dask_state['finished'])
        num_tasks = num_done + sum(len(dask_state[k])
                                   for k in ['ready', 'waiting', 'running'])
        if num_done < num_tasks:
            work_fraction = num_done / num_tasks if num_tasks > 0 else 0
            worked = work_fraction * self._total_work
            work = worked - self._last_worked
            if work > 0:
                _ProgressContext.instance().worked(work)
                self._last_worked = worked

    def _stop_thread(self):
        if self._running:
            self._running = False
            self._timer.join()
